/* Programa ROMEO.C - versao 1.00

   Instituto Tecnolgico de Aeronutica (ITA)
   Diviso de Engenharia Eletrnica
   Departamento de Sistemas e Controle

   Alunos: Gabriela Werner Gabriel (T. 03)
	   Ued Nei Seixas Andril  (T. 98)
	   Leonardo Alves Dias  (T. 98)
	   Fabio Eiji Yoshitome (T. 97)

   Orientadores: Prof. CAIRO L. NASCIMENTO JR.
		 Prof. EDUARDO HISASI YAGYU

   Descrio Sumria: Programa de Planejamento de Trajetria (A*) e
		 de Acionamento da Plaforma Mvel (ROMEO.C)
   Verso: 1.00
   Data: 20/Julho/2001
   Responsvel por esta verso: Cairo L. Nascimento Jr.

   Agradecimentos: Agradecemos ao Prof. Roberto H. Kawakami Galvao (ITA)
   por permitir o uso de sua implementacao do algoritmo A* em C.

   Descrio: Este programa usa o algoritmo A* para resolver o problema
   de navegao autnoma em um labirinto. A descrico do labirinto 
   fornecida em um arquivo .txt com sintaxe prpria. Tal labirinto no
   deve conter mais que MAX_NO ns por linha ou coluna (a constante MAX_NO
    definida durante a compilao do programa).

   Alm disso, o programa gera uma sequncia de comandos para a plataforma
   mvel seguir a trajetria planejada. Se desejado pelo usurio, tais
   comandos so convertidos em sinais que so enviados para a plataforma
   mvel, caso a placa de acionamento da plataforma esteja instalada no
   micro. Neste caso, este programa deve ento ser invocado usando
   como argumento a string definida em AC_PLAT.

   Nesta verso do programa, assume-se que o usurio  responsvel por
   verificar a presena da placa de acionamento da plataforma no micro.
   Nesta verso a sequncia de movimentos  enviada para o arquivo
   cujo nome  definido na string FileOutName.

   OBS: Compilado em TURBO C++ 3.0, small model, sem erros ou warnings.

*/

/* TO DO LIST:
- fazer fluxo de chamada das rotinas.
- verificar possveis "memory leaks".
- usar alocao dinmica de memria para reservar espao para
  as matrizes (usar malloc), ao invs de estabelecer um tamanho
  mximo para as dimenses das matrizes como agora.
- substituir na rotina reord_open o algoritmo "bubble sort" por
  algum outro algoritmo mais eficiente (ver livro H. Schildt, pag. 3).
- na ordem decrescente de importncia, minimizar o uso de:
    matrizes float, matrizes int, variveis float, variveis int.
- melhorar a entrada dos dados, p. ex. com menus, uso de um arquivo
  de configurao, admitir uma configurao padro e perguntar ao
  usurio se ele deseja modific-la.
- detectar automaticamente a velocidade do micro e ajustar o loop
  de delay, caso a funo delay no funcione.
- detectar automaticamente a presena da placa de acionamento da
  plataforma. Aparentemente se a placa no estiver no micro e se
  o programa tentar enviar comandos para a placa, o programa trava
  o micro.
- Melhorar o teste de erro das chamadas para OUTP pois pode ser
  que o teste atual no funcione, ou seja, para que depois de
  executada a instruo: value=outp(portnumber,outnumber)
  value seja = a outnumber mesmo se ocorreu erro na execuo da
  instruo outp.
- registrar a associao header files,funes usadas no programa,
  de forma tal que se as funes associadas com um header file forem
  retiradas poderemos retirar o header file da lista de "includes".
- talvez mostrar o driver name na tela se comear a aparecer problemas
  com micros que no vo para 640x480 VGA (ver livro do R. Lafore pag. 345).

Cuidados com C:
--------------
  - memory leak: o programa pode no dar erro de compilao ou de
	       execuo se for usado uma posio no definida
	       de um array, p. ex, definir int X[2] e usar X[2]=1; a=X[2]+1;
  - testes nos ifs: if (i=j) .... o correto  if (i==j)
  em ambos os casos o programa pode gerar resultados estranhos
  sem acusar erro de execuo ou de compilao
*/

#include <dos.h>      /* para delay(), sound(), nosound() */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>   /* para strcpy() */
#include <conio.h>    /* para cprintf(), outp()   */
#include <math.h>
#include <graphics.h> /* para as funes da tela grfica */

/* no de passos para deslocamento de 1 unidade de distncia horizontal
   no labirinto */
#define N_passos_desl_horz 269
/* no de passos para uma rotao de 360 graus da plataforma no labirinto */
#define N_passos_rot_360 570
/* delay (em ms) para o pulso de acionamento da plataforma */
#define DELAY_PULSO 7
/* argumento opcional na linha de comando (CASE INSENSITIVE) usado
   para acionar o hardware da plataforma, p. ex., se o nome do programa
   for romeo, ento para acionar a plataforma, o programa deve ser
   invocado como: "romeo ap" */
#define AC_PLAT "ap"
/* Nome do arquivo onde a sequencia de movimentos da plataforma
   sera' armazenada */
#define FileOutName "tabmov.txt"
/* dimenses das matrizes do labirinto e do algoritmo de planejamento
   da trajetria */
#define MAX_NO 20
#define MAX_NO_2 MAX_NO*MAX_NO
/* tamanho da string com as msgs da tela grfica */
#define BUFFER_SIZE 80
/* caractersticas dos fontes usados para as msgs na tela grfica */
#define FONT_TYPE DEFAULT_FONT
#define FONT_SIZE 1
#define FONT_COLOR WHITE        /* cor das msgs normais */
#define FONT_COLOR_WARNING YELLOW
#define FONT_COLOR_ERROR LIGHTRED    /* cor das msgs de erro */
/* variveis para a funo beep() */
#define BEEP_CONST 2 /* determina mtodo de gerao do beep pela rotina beep() */
#define BEEP_FREQ 750   /* frequncia do beep (mtodo 2) */
#define BEEP_DUR 200    /* durao do beep (mtodo 2)    */

#define ESC 27          /* cdigo ASCII (base decimal) para tecla ESC  */
#define PI 3.141592     /* usado para converso entre graus e radianos */


/* Prottipos das Rotinas Principais */
void ler_labirinto(void);
void ler_dados(void);
void inicializar_labirinto(void);
void desenhar_labirinto(void);
void planejar_trajetoria(void);
void tracar_trajetoria_final(void);
void gerar_mov_rot_transl(void);
void gerar_mov_trajetoria(void);
void inicializar_hard_plat(void);
void acionar_hard_plat(float rot_gr, float transl);

/* Prottipos das Rotinas Auxiliares */
float calch(int a, int b);
void expand(int N, float gn);
void reord_open(void);
void showpath(int N, int indcor);
void desenhar_seta(int i, int j, float tetag, int indcor);
void desenhar_circ(int indcor);
void testar_tecla_ESC(void);
void beep(void);
void preparar_topo_tg(int indcor);
void colocar_texto(int cod, int cor);

/* Declarao das variveis globais do programa */
/* Lembrete: varivel local tem precedncia sobre varivel global */
/* Variveis Principais */
int M[MAX_NO][MAX_NO],OPEN[MAX_NO_2][2],CLOSED[MAX_NO_2][2];
float GOPEN[MAX_NO_2],GCLOSED[MAX_NO_2],h[MAX_NO_2];
double rotac[MAX_NO_2],translac[MAX_NO_2];
int nclosed,nopen,nl,nc,is,js,ie,je,npt,rota[MAX_NO_2];
float xd,xv,xteta,teta0,custo_rota;
/* Variveis Auxiliares */
double mx,my,espacx,espacy,dx,dy,raio; /* estas variveis podem ser inteiras
porm para melhorar a preciso oo clculo dos movimentos da plataforma
 melhor defini-las como float ou double */
int fim_da_busca,ret_code,INI_PLAT_ERROR=1;
char visexp,pap,buffer[BUFFER_SIZE];
int cor[9]={ 15, 12, 2, 9, 9, 12, 14, 9, 9};
char cmd_ac_plat[20];

/* cor[.] --> tabela de cores em sequncia indicando:
	   (0-4) o tipo de n: livre, obstculo, inicial, final, em expanso
	   (5) as ligaes do n em expanso
	   (6) as ligaes dos ns que j foram expandidos
	   (7) a trajetria sendo considerada para o n em expanso
	   (8) a trajetria final
   Ex. de cores para VGA:
   DARK COLORS: (0 = black), (1 = blue), (2 = green), (3 = cyan),
   (4 = red), (5 = magenta), (6 = brown), (7 = lightgray)
   BRIGHT COLORS: (8 = darkgray), (9 = lightblue), (10 = lightgreen)
   (11 = lightcyan), (12 = lightred), (13 = lightmagenta),
   (14 = yellow), (15 = white)
   Obs: Para deixar o "source code" mais fcil de ser entendido,
	use nas funes de chamada de cor o nome da cor (em letras
	maisculas) ao invs do nmero correspondente.
*/

/*
Descrio das variveis globais (completar ...):
OPEN[.][0] --> N a expandir
OPEN[.][1] --> Precursor desse n
GOPEN[.] --> Custo da movimentao do n inicial at este n
CLOSED[.][0] --> N j expandido
CLOSED[.][1] --> Precursor desse n
GCLOSED[.] --> Custo da movimentao do n inicial at este n
h[.] --> Heuristica que mede a distancia direta entre o n [.] e o n final,
	ignorando os obstculos
xd --> custo do movimento na diagonal
xv --> custo do movimento na vertical
xteta --> custo de mudana no ngulo da plataforma
(xd, xv e xteta so normalizados em relao ao custo do movimento na horizontal)
teta0 e teta --> orientao inicial e atual da plataforma [em graus],
  se teta = 0 graus = posio horizontal
  se teta > 0 = sentido anti-horrio
  se teta < 0 = sentido horrio
nopen, nclosed --> numero de ns nas listas OPEN e CLOSED
nl,nc --> nmero de linhas e de colunas do labirinto
is,js --> linha e coluna do n inicial (Start)
ie,je --> linha e coluna do n final (End)
espacx,espacy --> espaamento entre os ns no modo grfico
raio --> raio usado para traar os ns como crculos
cor[.] --> tabela de cores para tipos de ns, ligaes e trajetrias,
	   conforme indicado acima.
fim_da_busca --> flag que indica se o algoritmo encontrou o n final (flag=1)
visexp, pap --> opes do usurio
	visexp SIM --> Visualizar a expanso dos ns
	pap SIM --> Visualizar a expanso dos ns passo-a-passo
*/

/*------------------------------------------------------------------------*/
/* Rotina principal por onde o programa  iniciado e deve terminar em
   condies normais. */
void main(int argc, char *argv[])
{

  /* 1 - Entrada de dados:
     a) ler o arquivo que contm a descrio do labirinto
     b) pedir ao usurio os dados de entrada:
	- se o usurio deseja visualizar a expanso dos ns,
	- se SIM, se a expanso dos ns deve ser passo a passo,
	- valor de xd (custo do movimento na diagonal).
  */
  ler_labirinto();
  ler_dados();

  /* 2 - Desenhar o labirinto na tela:
     a) Inicializar o labirinto:
	- calcular para cada n o custo heurstico entre esse n e
	o n final (matriz h),
     b) Desenhar o labirinto no modo grfico.
     Assume-se que os ns do labirinto so de 0 a nc*nl-1, sendo que
     o n (i,j)  numerado como nc*i+j.
  */
  inicializar_labirinto();
  desenhar_labirinto();

  /* 3 - Planejar e traar a trajetria:
  */
  planejar_trajetoria();
  tracar_trajetoria_final();

  /* 4 - Gerar os movimentos de rotao e translao para cada ponto
	 da trajetria e acionar o hardware da plataforma se desejado.
  */
  if (ret_code==0)
     {
     gerar_mov_rot_transl();
	testar_tecla_ESC();  /* PARADA */
	/* se a linha de comando tiver um argumento, guardar tal argumento */
	if (argc==2) strcpy(cmd_ac_plat,argv[1]);
	else strcpy(cmd_ac_plat,"nao_acionar");
     gerar_mov_trajetoria();
	/* fechar programa e retornar ao sistema operacional */
	closegraph();
	textcolor(YELLOW);
	cprintf("\r\n   Bye bye ...");
	textcolor(WHITE); normvideo(); /* select normal intensity characteres */
	cprintf("\r\n");  /* imprime estes caracteres para voltar o texto ao modo
			     normal (branco com intensidade normal) */
     exit(0);
     }
  else
     {
     closegraph();
     textcolor(FONT_COLOR_ERROR);
     cprintf("\r\n\r\n ERRO: No foi possvel encontrar uma trajetria.");
     cprintf("    \r\n       Verifique o labirinto.");
     textcolor(WHITE); normvideo(); /* select normal intensity characteres */
     cprintf("\r\n");  /* imprime estes caracteres para voltar o texto ao modo
			  normal (branco com intensidade normal) */
     exit(1);
     }

} /* FIM DO PROGRAMA PRINCIPAL */
/*------------------------------------------------------------------------*/
/* Ler o arquivo-texto que contm a descrio do labirinto */
void ler_labirinto()
{
  int i,j,noi,nof,nerros;      /* noi,nof sao respectivamente          */
  FILE *in;                    /* o numero de nos iniciais e de finais */
  char arquivo[30];

  /* clrscr(); */
  printf("\nROMEO 20-Jul-2001\nArquivo (nome+extenso) com a descrio do labirinto (ex. lab1.txt): ");
  scanf("%s",arquivo);
  if ((in = fopen(arquivo, "rt")) == NULL)   /* arquivo no existente */
      {
      beep();  /* BEEP */
      fprintf(stderr,"\n ERRO: No  possvel abrir o arquivo com a descrio do labirinto.");
      fprintf(stderr,"\n       Verifique o nome do arquivo e se o arquivo existe. \n\n");
      exit(1);
      }
  fscanf(in,"%d %d",&nl,&nc);  /* ler nmero de linhas e colunas do labirinto */
  noi=0; nof=0;
  nerros=0;
  for (i=0;i<nl;i++)
	for (j=0;j<nc;j++)
	{
	  fscanf(in,"%d",&M[i][j]);
	  if ((M[i][j]!=0) && (M[i][j]!=1) && (M[i][j]!=2) && (M[i][j]!=3))
		{
		nerros++;
		fprintf(stderr,"\n ERRO: Labirinto invlido, Posio: %d, %d; Valor: %d",i+1,j+1,M[i][j]);
		}
	  if (M[i][j]==2)      /* N inicial */
		{ noi++; is=i; js=j; }
	  if (M[i][j]==3)      /* N final */
		{ nof++; ie=i; je=j; }
	}
  /* fim do loop for */

  if (noi!=1)
	{
	nerros++;
	fprintf(stderr,"\n ERRO: Nmero incorreto de ns iniciais: %d", noi);
	}
  if (nof!=1)
	{
	nerros++;
	fprintf(stderr,"\n ERRO: Nmero incorreto de ns finais: %d", nof);
	}

  /* Fechar o arquivo aberto */
  fclose(in);

  /* Verificar nmero de erros detectados */
  if (nerros!=0)
	{
	beep();  /* BEEP */
	fprintf(stderr, "\n\n ERRO: Labirinto no vlido. Redefinir labirinto.\n\n");
	exit(1);
	}

} /* FIM DA ROTINA */
/*------------------------------------------------------------------------*/
/* Pedir ao usurio os dados de entrada:
     - se o usurio deseja visualizar a expanso dos ns
     - se SIM, se a expanso dos ns deve ser passo a passo
     - xd (custo do movimento na diagonal)
*/
void ler_dados()
{

  printf("Deseja visualizar a expanso dos ns? (S/N): ");
  do
    {
    visexp=getch();
    if ((visexp=='s') || (visexp=='S') || (visexp=='n') || (visexp=='N'))
       printf("%c",visexp);
    }
  while ((visexp!='s') && (visexp!='S') && (visexp!='n') && (visexp!='N'));

  if ((visexp=='s') || (visexp=='S'))
     {
     printf("\nDeseja visualizar a expanso passo a passo? (S/N): ");
     do
       {
       pap=getch();
       if ((pap=='s') || (pap=='S') || (pap=='n') || (pap=='N'))
	  printf("%c",pap);
       }
     while ((pap!='s') && (pap!='S') && (pap!='n') && (pap!='N'));
     }
  else
     pap='n';

  xv=-1.0;
  while (xv <= 0.0)
     {
     printf("\nEntre com o valor de XV - custo do movimento na vertical (ex. 1): ");
     scanf("%f",&xv);
     if (xv <= 0.0)
	{
	beep();  /* BEEP */
	fprintf(stderr, "\n XV no vlido. XV deve ser > 0. Tente de novo ...\n");
	}
     }

  xteta=-1.0;
  while (xteta < 0.0)
     {
     printf("Entre com o valor de XTETA - custo para rotao de 90 graus (ex. 1): ");
     scanf("%f",&xteta);
     if (xteta < 0.0)
	{
	beep();  /* BEEP */
	fprintf(stderr, "\n XTETA no vlido. XTETA deve ser >= 0. Tente de novo ...\n");
	}
     }

  teta0=200.0;
  while ((teta0 < -180.0) || (teta0 > 180.0))
     {
     printf("Entre com a orientao inicial TETA0 em graus (ex. -45): ");
     scanf("%f",&teta0);
     if ((teta0 < -180.0) || (teta0 > 180.0))
	{
	beep();  /* BEEP */
	fprintf(stderr, "\n TETA0 no vlido. Deve ser -180 <= TETA0 <= 180. Tente de novo ...\n");
	}
     }

} /* FIM DA ROTINA */
/*------------------------------------------------------------------------*/
/* Inicializar o labirinto:
     - calcular para cada n o custo heurstico entre esse n e o n final,
       ou seja, definir os valores da matriz h.
  Assume-se que os ns do labirinto de so de 0 a nc*nl-1, sendo que
       o n (i,j)  numerado como nc*i+j.
  Esta rotina chama a rotina calch().
*/
void inicializar_labirinto()
{
int i,j;

/* calcular para cada n o custo heurstico entre esse n e o n final */
for (i=0;i<nl;i++)
	for (j=0;j<nc;j++)
		h[nc*i+j]=calch(i,j);

} /* FIM DA ROTINA */
/*------------------------------------------------------------------------*/
/* Calcular o custo do deslocamento direto entre cada n do labirinto e
   o n final, ignorando os obstculos.
   Esta rotina  chamada pela funo inicializar_labirinto().
*/
float calch(int i, int j)
{
  float dist_hor,dist_vert,H=0;

  dist_hor=abs(je-j);
  dist_vert=xv*abs(ie-i);
  H=sqrt(dist_hor*dist_hor + dist_vert*dist_vert);
  return(H);

} /* FIM DA ROTINA */
/*------------------------------------------------------------------------*/
/* Desenhar o labirinto no modo grfico */
void desenhar_labirinto()
{
   int i,j,indcor,espymax;
   int gdriver = DETECT, gmode, errorcode, mode;
   /* mx,my --> dimensoes da tela

   /* initialize graphics mode */
   initgraph(&gdriver, &gmode, "");

   /* read result of initialization */
   errorcode = graphresult();

   if (errorcode != grOk)  /* an error occurred */
      {
      beep(); /* BEEP */
      fprintf(stderr,"Graphics error: %s\n", grapherrormsg(errorcode));
      fprintf(stderr,"Press any key to halt the program:");
      getch();
      exit(1);             /* return with error code */
      }

   /* Obter as mximas dimenses da tela */
   mx=getmaxx()+1;
   my=getmaxy()+1;

   /* Calcula o espaamento necessrio entre os ns na tela,
   assumindo inicialmente que se o grfico for feito proporcionalmente
   o espao disponvel em y  maior que o espao necessrio.
   em X: mx = dx+(nc-1)*espacx+dx, onde espacx/2 --> espacx=mx/2
   em Y: my = dy+(nl-1)*espacy+dy  --> dy=(my-(nl-1)*espacy)/2 */
   espacx=mx/nc;
   espacy=xv*((float)(mx/nc));
   dx=espacx/2;      /* a folga lateral em x e em y */
   dy=(my-(nl-1)*((xv*mx)/nc))/2;

   /* verifica se o espaamento necessrio para os ns em Y  maior
   que o mximo espao disponvel. Se SIM, redefinir os espaamentos
   em X e Y e as folgas dx e dy. */
   espymax=my/nl;

   if (espacy > espymax)
      {
      espacx=espacx*(((float)(espymax))/((float)(espacy)));
      espacy=espymax;
      dy=((float)(my/nl))/2;
      dx=(mx-(nc-1)*espacx)/2;
      }

   /* definir o raio dos crculos */
   raio=espacx/4;
   if (espacy<espacx) raio=espacy/4;

   /* traar os crculos dos ns, sendo a cor de cada n de acordo
   com o seu tipo */
   for (i=0;i<nl;i++)
       for (j=0;j<nc;j++)
	   {
	   indcor=M[i][j];
	   setcolor(cor[indcor]);
	   circle(dx+j*espacx,dy+i*espacy,raio);
	   /* Preencher o crculo no caso da posio (i,j) no ser n livre */
	   if (M[i][j]!=0)
	      {
	      setfillstyle(SOLID_FILL,cor[indcor]);
	      floodfill(dx+j*espacx,dy+i*espacy,cor[indcor]);
	      }
	   } /* fim dos 2 loops for */

   /* Desenhar a seta com a orientao inicial TETA0 */
   desenhar_seta(dx+js*espacx,dy+is*espacy,teta0,cor[2]);

   /* Mostrar o nome do modo grfico */
   mode = getgraphmode();
   preparar_topo_tg(FONT_COLOR);
   sprintf(buffer, "%s  o modo grfico em uso",getmodename(mode));
   outtext(buffer);
   testar_tecla_ESC();

} /* FIM DA ROTINA */
/*------------------------------------------------------------------------*/
/* Planejamento da trajetoria:
   Esta rotina chama as rotinas expand() e reord_open()
*/
void planejar_trajetoria()
{
int i,N;
float gn;

/* Os ns so numerados de 0 a nc*nl-1
   O n (i,j)  numerado como nc*i+j
   nl,nc --> nmero de linhas e de colunas do labirinto
   is,js --> linha e coluna do n inicial (Start)
   ie,je --> linha e coluna do n final   (End)
   nopen, nclosed --> numero de ns nas listas OPEN e CLOSED
   fim_da_busca --> flag que indica se o algoritmo encontrou o n final (flag=1)

   Matrizes e vetores:
   OPEN[.][0] --> N a expandir
   OPEN[.][1] --> Precursor desse n
   GOPEN[.] --> Custo da movimentao do n inicial at este n
   CLOSED[.][0] --> N j expandido
   CLOSED[.][1] --> Precursor desse n
   GCLOSED[.] --> Custo da movimentao do n inicial at este n
   h[.] --> Heurstica que mede a distancia direta entre o n [.]
	    e o n final, ignorando os obstculos
*/

/* Clculo de xd (custo do movimento na diagonal) a partir de:
   xh = custo de movimento na horizontal, assumido = 1; e
   xv = custo de movimento na vertical */
  xd=sqrt(1+xv*xv);

/* Inicializar OPEN com o n inicial = n de partida */
  OPEN[0][0]=nc*is+js;    /* is,js = linha e coluna do n inicial (Start) */
  OPEN[0][1]=-1;          /* Antecessor do n inicial = -1 */
  GOPEN[0]=0;             /* Custo Inicial g = 0 */
  nopen=1;
  nclosed=0;
  fim_da_busca=0;
  npt=1;  /* nmero de pontos na trajetria */

  do
  {
    /* Toma o primeiro elemento de OPEN, chamando-o de N */
	N=OPEN[0][0];
	gn=GOPEN[0];     /* Toma o seu custo gn do n N */

    /* Retira esse elemento de OPEN e o coloca em CLOSED */
	CLOSED[nclosed][0]=N;
	CLOSED[nclosed][1]=OPEN[0][1];
	GCLOSED[nclosed]=GOPEN[0];
	nclosed++;
	nopen--;
    /* Desloca os elementos da matriz OPEN e do vetor GOPEN visto que
       o primeiro elemento de OPEN foi transferido para CLOSED */
	for (i=0;i<nopen;i++)
	{
	  OPEN[i][0]=OPEN[i+1][0];
	  OPEN[i][1]=OPEN[i+1][1];
	  GOPEN[i]=GOPEN[i+1];
	}

    /* Se N ainda no  o n final, N  expandido */
	if (N!=ie*nc+je)
	{
	  expand(N,gn);
	  reord_open();     /* Reordena OPEN com base no custo f'=g+h */
	}
  }
  /* Encerra a busca:
       - se o n final foi encontrado, OU
       - se no h mais nehum n para serem expandido.
     Assim o loop continua sendo executado enquanto:
       - o n final ainda no foi encontrado, E
       - existem ns para serem expandidos (nopen>0). */
  while ((nopen>0) && (fim_da_busca!=1));

  /* Se ocorreu o fim da busca ento colocar o n final na lista
  CLOSED sendo seu precursor o n N
  Tambm deve-se aumentar nclosed de 1 unidade neste caso para
  indicar a presenca de mais 1 elemento na lista CLOSED */
  if (fim_da_busca==1)
     {
     CLOSED[nclosed][1]=N;
     CLOSED[nclosed][0]=ie*nc+je;
     nclosed++;
     }

} /* FIM DA ROTINA */
/*------------------------------------------------------------*/
/* Rotina que implementa a expanso do n N
   Esta rotina  chamada pela rotina planejar_trajetoria()
   Esta rotina chama a rotina desenhar_seta() e desenhar_circ()

   Ao ser expandido um n, cada n sucessor  analisado e se
   o caminho ento sendo considerado do n de partida at o n sucessor
   for melhor (igual ou pior) do que o caminho armazenado para tal
   n ser desenhado um crculo no canto superior direito verde (vermelho).
   Se o caminho for novo o crculo ser de cor branca.
*/
void expand(int N, float gn)
{
  int in,jn,Np,ip,jp,i,j,a,suc,indcor,novo;
  float gsuc;
  float tetap,tetac,dteta;

  in=div(N,nc).quot;  /* in,jn = posio do n N na grade */
  jn=div(N,nc).rem;

  Np=CLOSED[nclosed-1][1];  /* recupera o pai do n N = Np*/
  if (M[in][jn]==2)
     { ip=-1; jp=-1; } /* ip,jp = posio do n Np na grade */
  else
     { ip=div(Np,nc).quot; jp=div(Np,nc).rem; }

  /* Cada n tem:
     - o seu ngulo de chegada no n (tetac) e
     - o seu ngulo de partida do n (tetap)
     Em um trajetria de n A para n B, temos:
       tetap(n A) = tetac(n B)
       dteta(n X) = tetap(n X) - tetac(n X)   */
  /* Determine o ngulo de chegada no n que vai ser expandido */
     if (M[in][jn]==2) tetac=teta0; /* se for n de partida */
     else tetac=atan2(-(in-ip)*espacy,(jn-jp)*espacx)*180.0/PI;
     /* atan2(y,x) retorna um ngulo no intervalo [-pi,pi] */

  /* indicar na tela o n que est sendo expandido e a trajetria
     at ele escolhida at agora pelo algoritmo. Porm se for o
     n inicial (de partida) tais indicaes no so necessrias */
  if (M[in][jn]!=2)
     {
     setcolor(cor[4]);
     circle(dx+jn*espacx,dy+in*espacy,raio);
     }

  if ((pap=='s') || (pap=='S'))
     {
     if (M[in][jn]!=2) showpath(in*nc+jn,cor[7]);
     /* preparar o topo da tela grfica */
     preparar_topo_tg(FONT_COLOR);
     /* indicar na tela grfica os valores de g, h e g+h para este n */
     sprintf(buffer,"g = %f, h = %f, g+h = %f, NPT = %d",GCLOSED[nclosed-1],h[N],GCLOSED[nclosed-1]+h[N],npt);
     outtext(buffer);
     testar_tecla_ESC();
     }

  /* Loop em i e j para gerar novos ns a partir do n N */
  /* i,j= coordenadas do novo n */
  for (i=in-1;i<=in+1;i++)
	for (j=jn-1;j<=jn+1;j++)
	    {
	    suc=nc*i+j;
	    // printf("\n %d %d %d %d",in, jn, i,j); */
	    // printf("\n N=%d Pai de N=%d, suc=%d",N,CLOSED[nclosed-1][1],suc);
	    /* regra de validade do novo n */
	    if ( (i>=0) && (j>=0) && (i<nl) && (j<nc) && (M[i][j]!=1) && (suc!=N) && (CLOSED[nclosed-1][1]!=suc) && (fim_da_busca==0))
	       {
		 /* Mostra a expanso do n no video */
		 if ((visexp=='s') || (visexp=='S'))
		    {
		    setcolor(cor[5]);
		    line(dx+jn*espacx,dy+in*espacy,dx+j*espacx,dy+i*espacy);
		    }

		 /* Calcular o ngulo de partida tetap considerando o
		    movimento de (in,jn) para (i,j) e que estamos em
		    (in,jn) */
		 if (i==in)
		    {
		    if (j>=jn) tetap=0;
		    else tetap=180;
		    }
		 else if (j==jn)
		    {
		    if (i<=in) tetap=90;
		    else tetap=-90;
		    }
		 else
		    tetap=atan2(-(i-in)*espacy,(j-jn)*espacx)*180.0/PI;

		 if (fabs(tetap)>180.1 || fabs(tetac)>180.1)
		 /* 180.1  usado no lugar de 180 para levar em conta
		 a impreciso da funo atan2 quando atribuda  varivel
		 tipo float */
		    {
		    beep();  /* BEEP */
		    closegraph();
		    fprintf(stderr,"\n\n ERRO no valor de tetac (%f) ou tetap (%f) !!! \n\n",tetac,tetap);
		    exit(1);
		    }
		 /* assumindo que neste ponto tetac e tetap esto no
		    intervalo [-180,180] graus */
		 dteta=fabs(tetac-tetap);
		 while (dteta>180.0) dteta=fabs(dteta-360.0);
		 /* */
		 //preparar_topo_tg(cor[0]);
		 //sprintf(buffer,"tetac = %f,tetap = %f, dteta = %f",tetac,tetap,dteta);
		 //outtext(buffer);
		 /* */

		 /* se n sucessor est na mesma linha de N */
		 if ((i==in) && (j!=jn)) gsuc=gn+1;
		 /* se n sucessor est na mesma coluna de N */
		 if ((j==jn) && (i!=in)) gsuc=gn+xv;
		 /* se n sucessor no na mesma linha nem na mesma coluna de N */
		 if ((i!=in) && (j!=jn)) gsuc=gn+xd;

		 gsuc=gsuc+xteta*dteta/90.0;

		 /* se n sucessor est na mesma linha e na mesma coluna de N */
		 if ((i==in) && (j==jn))
		    {
		    beep();  /* BEEP */
		    /* preparar topo da tela grfica */
		    preparar_topo_tg(FONT_COLOR_ERROR);
		    outtext("ERRO !! N sucessor = N em expanso ... Verifique algoritmo ...");
		    getch();
		    closegraph();
		    fprintf(stderr,"\n\n ERRO!! N sucessor = N em expanso ...");
		    fprintf(stderr,  "\n       Verifique o algoritmo de expanso dos ns ....\n\n");
		    exit(1);
		    }

		 /* se o n sucessor for o n final ento indicar com o
		   flag fim_da_busca=1 e armazenar o valor de gn como sendo
		   o custo total da rota */
		 if ((i==ie) && (j==je))
		    {
		    fim_da_busca=1;       /* O n final foi atingido */
		    custo_rota=gsuc;      /* custo toal da rota      */
		    }

		 novo=1;

		 /* Se o sucessor j se encontra na lista OPEN ou na lista CLOSED,
		 o programa chama a cpia do sucessor em OPEN e verifica se o custo
		 g da copia  maior. Em caso afirmativo, substitui-se a cpia por suc.
		 Em caso negativo, descarta-se suc. O mesmo procedimento  repetido
		 com CLOSED */

		 for (a=0;a<nopen;a++)
		     if (suc==OPEN[a][0])
			{
			novo=0;
			if (gsuc<GOPEN[a])
			   {
			   GOPEN[a]=gsuc;
			   OPEN[a][1]=N;
			   desenhar_circ(cor[2]); /* cor verde */
			   }
			else
			   {
			   gsuc=GOPEN[a];
			   desenhar_circ(cor[1]); /* cor vermelha */
			   }
			} /* end do for e do if */

		 for (a=0;a<nclosed;a++)
		     if (suc==CLOSED[a][0])
			{
			novo=0;
			if (gsuc<GCLOSED[a])
			   {
			   GCLOSED[a]=gsuc;
			   CLOSED[a][1]=N;
			   desenhar_circ(cor[2]); /* cor verde */
			   }
			else
			   {
			   gsuc=GCLOSED[a];
			   desenhar_circ(cor[1]); /* cor vermelha */
			   }
			} /* end do for e do if */

		 /* Se suc nao se encontra em OPEN nem em CLOSED, ele eh adicionado
		 a OPEN */
		 if (novo==1)
		    {
		    OPEN[nopen][0]=suc;
		    GOPEN[nopen]=gsuc;
		    OPEN[nopen][1]=N;
		    nopen++;
		    desenhar_circ(cor[0]); /* cor branca */
		    } /* end do if */

		 if ((pap=='s') || (pap=='S'))
		    {
		    /* preparar topo da tela grfica */
		    preparar_topo_tg(cor[6]);
		    /* indicar na tela grfica os valores de g, h e g+h
		    para este n  */
		    sprintf(buffer,"g = %f, h = %f, g+h = %f",gsuc,h[suc],gsuc+h[suc]);
		    outtext(buffer);
		    testar_tecla_ESC();
		    }

	       } /* end do if que testa a validade do n */
	    } /* end dos 2 for's */

/* colocar a cor original no n que foi expandido. Porm se o n
   expandido era o n inicial (de partida) isto no  necessrio.
   Retraar a trajetria at este n na cor para as ligaes dos ns
   que j foram expandidos para preparar a tela para o prximo n */
  //if ((pap=='s') | (pap=='S')) getch(); /* parada depois da expansao deste n */
  if (M[in][jn]!=2)
     {
     indcor=M[in][jn];
     setcolor(cor[indcor]);
     circle(dx+jn*espacx,dy+in*espacy,raio);
     if ((pap=='s') || (pap=='S')) showpath(in*nc+jn,cor[6]);
     }

/* se foi pedido para ver a expanso dos ns ento retraar as linhas entre
   o n que foi expandido e seus sucessores, para preparar a tela
   para a expanso do prximo n */
   if ((visexp=='s') || (visexp=='S'))
      for (i=in-1;i<=in+1;i++)
	  for (j=jn-1;j<=jn+1;j++)
	      {
	      suc=nc*i+j;
	      /* regra de validade do novo n */
	      /* este teste no inclui (fim_da_busca==0) como antes para que
	      a correo da cor acontea tambm para o ltimo passo antes de
	      atingir o n final  */
	      if ( (i>=0) && (j>=0) && (i<nl) && (j<nc) && (M[i][j]!=1) && (suc!=N) && (CLOSED[nclosed-1][1]!=suc))
		 {
		 /* Mostrar as ligaes deste n como sendo
		    de um n que j foi expandido */
		 setcolor(cor[6]);
		 line(dx+jn*espacx,dy+in*espacy,dx+j*espacx,dy+i*espacy);
		 } /* end do if */
	      }  /* end dos 2 for's e do if */

   /* apagar crculo no canto superior direito */
   desenhar_circ(BLACK);  /* cor preta */

} /* FIM DA ROTINA */
/*------------------------------------------------------------------------*/
/* Reordena a lista OPEN com base no custo f'=f+g de cada um de seus elementos.
   A reordenao  feita segundo a tcnica de "bubble sort".
   Esta rotina  chamada pela rotina planejar_trajetoria() */
void reord_open()
{
  int N,flag,a,b,c,TEMP1[MAX_NO_2][2];
  float fn,TEMP2[MAX_NO_2];

  TEMP1[0][0]=OPEN[0][0];
  TEMP1[0][1]=OPEN[0][1];
  TEMP2[0]=GOPEN[0];
  for (a=1;a<nopen;a++)
  {
	N=OPEN[a][0];
	fn=GOPEN[a]+h[N];
	flag=0;
	for (b=0;b<a;b++)
	  if (fn<TEMP2[b]+h[TEMP1[b][0]])
	  {
		for (c=a;c>b;c--)
		{
		  TEMP1[c][0]=TEMP1[c-1][0];
		  TEMP1[c][1]=TEMP1[c-1][1];
		  TEMP2[c]=TEMP2[c-1];
		}
		TEMP1[b][0]=N;
		TEMP1[b][1]=OPEN[a][1];
		TEMP2[b]=GOPEN[a];
		flag=1;
		b=a;
	  }
	if (flag==0)
	{
	  TEMP1[a][0]=N;
	  TEMP1[a][1]=OPEN[a][1];
	  TEMP2[a]=GOPEN[a];
	}
  }
  for (a=0;a<nopen;a++)
  {
	OPEN[a][0]=TEMP1[a][0];
	OPEN[a][1]=TEMP1[a][1];
	GOPEN[a]=TEMP2[a];
  }

}  /* FIM DA ROTINA */
/*------------------------------------------------------------------------*/
/* Desenha a rota encontrada na tela desde o ponto de partida at
   o n em expanso.
   Como na lista CLOSED cada n tem apenas 1 precursor, ento a
   rota  recuperada no sentido reverso, ou seja, do n N at o n inicial.*/
void showpath(int N, int indcor)
{
  int iN,jN,NP,iNP,jNP,i;

  setcolor(indcor);
  iN=div(N,nc).quot;
  jN=div(N,nc).rem;
  moveto(dx+jN*espacx,dy+iN*espacy);
  rota[0]=N;
  i=0;
  npt=1; /* nmero de pontos na trajetria */
  /* Se esta for a trajetria final usar linha com THICK_WIDTH */
  if ((iN==ie) && (jN==je))
     setlinestyle(SOLID_LINE,0,THICK_WIDTH);
  /* Percorre os ponteiros de CLOSED do n final ate o n inicial */
  do
  {
     /* testar o valor de i = indice dos elementos de CLOSED */
     if (i<0 || i>=nclosed)
	 {
	 beep();  /* BEEP */
	 /* preparar topo da tela grfica */
	 preparar_topo_tg(FONT_COLOR_ERROR);
	 outtext("ERRO !! ndice no vlido na matriz NCLOSED, rotina SHOWPATH ...");
	 getch();
	 closegraph();
	 fprintf(stderr,"\n\nERRO !! ndice no vlido na matriz NCLOSED, rotina SHOWPATH ... i = %d\n\n",i);
	 exit(1);
	 }
     else if (CLOSED[i][0]==N)
	{
	NP=CLOSED[i][1]; /* NP = Next Point */
	iNP=div(NP,nc).quot;
	jNP=div(NP,nc).rem;
	lineto(dx+jNP*espacx,dy+iNP*espacy);
	rota[npt]=NP;
	/* preparar variveis do loop para novo ponto */
	N=NP;
	npt=npt+1;
	i=-1; /* quando o if for executado de novo i=0 */
	}
     i++;
  }
  while (N!=(is*nc+js));

}
/*------------------------------------------------------------------------*/
void tracar_trajetoria_final()
{
  //int i;

  /* Se o n final foi encontrado, mostrar a rota escolhida */
  if (fim_da_busca==1)
     {
     showpath(ie*nc+je,cor[8]);  /* altera a cor e a posio do texto */
     /* preparar topo da tela grfica */
     preparar_topo_tg(FONT_COLOR);
     sprintf(buffer,"Trajetria encontrada com custo total = %f e NPT = %d",custo_rota,npt);
     outtext(buffer);
     ret_code=0;
     }
  else
     {
     beep();  beep(); beep(); /* BEEP */
     /* preparar topo da tela grfica */
     preparar_topo_tg(FONT_COLOR_ERROR);
     outtext("ERRO !! Trajetria no encontrada ... Verifique labirinto ...");
     getch(); /* PARADA */
     preparar_topo_tg(FONT_COLOR);
     outtext("FIM PREMATURO ... Pressione ESC para finalizar ...");
     setcolor(WHITE);
     settextjustify(RIGHT_TEXT,TOP_TEXT);
     moveto(mx,dy+(nl-1)*espacy);
     outtext("ESC");
     while (getch() != ESC);

     ret_code=1;  /* Se no foi encontrada uma rota, retornar ao programa
		     principal com ret_code DIF 0 (ERRO) */
     }

     //printf("\n\n   Trajetria encontrada com custo total = %f e NPT= %d\n",custo_rota,npt);
     //for (i=0;i<npt;i++) printf(" %d",rota[i]);
     //for (i=npt-1;i>=0;i--) printf(" %d",rota[i]);
     //ret_code=0;       /* Finaliza o programa retornando 0 (OK) */

}  /* FIM DA ROTINA */
/*------------------------------------------------------------------------*/
/* Uma vez determinados os ns que fazem parte da rota e sabendo-se a
   orientao inicial da plataforma esta rotina gera os comandos em
   rotao e deslocamento para cada n da rota */
void gerar_mov_rot_transl()
{
   int ind,ip,jp,ic,jc;

   /* Incluindo os pontos inicial e final a rota escolhida tem npt pontos,
   ou seja, rota[0] = N final e rota[npt-1] = n inicial
   Neste ponto do programa npt >= 2 porque se foi determinada uma trajetria
   vlida, tal trajetria tem pelo menos 2 pontos dado que o n de partida
   tem que ser diferente do n de chegada */

   /* calcular o ngulo de chegada em cada n da rota e armazen-los
      no vetor rotac. Estamos assumindo que o segmento da rota 
      de (ip,jp) --> (ic,jc) e estamos em (ic,jc) */
   rotac[0]=teta0;
   ip=is;
   jp=js;
   for (ind=1;ind<npt;ind++)
       {
       ic=div(rota[npt-1-ind],nc).quot;
       jc=div(rota[npt-1-ind],nc).rem;

       if (ic==ip) /* mesma horizontal */
	  {
	  translac[ind-1]=1;
	  if (jp<=jc) rotac[ind]=0;
	  else rotac[ind]=180;
	  }
       else if (jc==jp) /* mesma vertical */
	  {
	  translac[ind-1]=xv;
	  if (ic<=ip) rotac[ind]=90;
	  else rotac[ind]=-90;
	  }
       else   /* movimento na diagonal */
	  {
	  translac[ind-1]=xd;
	  rotac[ind]=atan2(-(ic-ip)*espacy,(jc-jp)*espacx)*((double)(180.0/PI));
	  }

       if (fabs(rotac[ind])>180.1 || ((ic==ip)&&(jc==jp)))
	  {
	  beep(); /* BEEP */
	  getch();
	  closegraph();
	  fprintf(stderr,"ERRO !!! na rotina gera_traj_rd !! \n\n");
	  exit(1);
	  }

       ip=ic;
       jp=jc;

       } /* fim do loop FOR*/

   /* Como (ngulo de partida do n I) = (ngulo de chegada no n I+1)
      e (movimento angular no n I) = (ngulo de partida do n I) - (ngulo de chegada no n I)
	(movimento angular no n I) = (ngulo de chegada no n I) - (ngulo de chegada no n I+1) */
   for (ind=0;ind<npt-1;ind++)
       {
       rotac[ind]=rotac[ind+1]-rotac[ind];
       /* colocando rotac[ind] no intervalo [-180,180] */
       while (rotac[ind]>180.0)  rotac[ind]=rotac[ind]-360.0;
       while (rotac[ind]<-180.0) rotac[ind]=rotac[ind]+360.0;
       }
   /* Assume-se que ao chegar no ponto final nenhum ajuste  feito na
      posio angular da plataforma */
   rotac[npt-1]=0; /* definindo movimento angular para o n de chegada */
   translac[npt-1]=0; /* definindo movimento de translao para o n de chegada */

} /* FIM DA ROTINA */
/*------------------------------------------------------------------------*/
void gerar_mov_trajetoria()
{
   int ind,CodEscrita=1;
   double xp,yp,xc,yc,teta;
   FILE *foutptr;

   /* Msg avisando que vai comear a enviar os comandos para a plataforma */
   preparar_topo_tg(FONT_COLOR);
   outtext("Incio da rotina de envio de comandos para a plataforma");
   inicializar_hard_plat();

   /* Usando apenas os vetores rotac e desloc refazer a trajetria */
   setlinestyle(SOLID_LINE,0,THICK_WIDTH); /* usar NORM_WIDTH ou THICK_WIDTH*/
   xp=dx+js*espacx;
   yp=dy+is*espacy;
   teta=teta0;
   desenhar_seta(xp,yp,teta0,cor[2]);
   testar_tecla_ESC(); /* PARADA */

   /* Criar arquivo de saida para gravar a tabela de movimentos      */
   /* Se ja' existe o arquivo de saida, entao a tabela de movimentos */
   /* nao sera' criada e uma msg de erro e' emitida para o usuario   */
   if ((foutptr = fopen(FileOutName, "r")) != NULL)
      {
      CodEscrita=0;
      beep();  /* BEEP */
      preparar_topo_tg(FONT_COLOR_WARNING);
      outtext("Arquivo de saida ja' existe. Nova tabela de movimentos nao sera' criada.");
      fclose(foutptr);
      getch();
      }
      else if ((foutptr = fopen(FileOutName, "wt")) == NULL)
      /* Erro se arquivo de saida nao existe e nao foi possivel abri-lo. */
      /* Este teste nao aponta erro se o arquivo de saida ja' existe,    */
      /* mesmo se o disco esta' cheio ou protegido contra escrita.       */
      {
      CodEscrita=0;
      beep();  /* BEEP */
      preparar_topo_tg(FONT_COLOR_WARNING);
      outtext("Nao foi possivel criar arquivo de saida com a tabela de movimentos.");
      getch();
      preparar_topo_tg(FONT_COLOR_WARNING);
      outtext("Possveis causas: disco cheio ou protegido contra escrita.");
      getch();
      }

   /* Loop para executar cada a sequncia de movimentos */
   for (ind=0;ind<npt-1;ind++)
       {
	 /* Mostrar msg para usurio */
	 preparar_topo_tg(FONT_COLOR); /* texto na cor branca */
	 sprintf(buffer,"Mov: %d --> teta = %.2f, desl = %.3f",ind+1,rotac[ind],translac[ind]);
	 outtext(buffer);

	/* escrever no arquivo de saida se ele estiver aberto para escrita */
	if (CodEscrita == 1)
	   fprintf(foutptr,"%d: %g, %g\n",ind+1,rotac[ind],translac[ind]);

	testar_tecla_ESC();
	/* Calcular novo ngulo */
	teta=teta+rotac[ind];
	/* Desenhar seta com o novo ngulo no ponto de partida */
	desenhar_seta(xp,yp,teta,cor[2]);
	colocar_texto(1,YELLOW); /* escrever "rotao" */

       /* acionar hardware da plataforma para fazer movimento de rotao */
       acionar_hard_plat(rotac[ind],0);
	 testar_tecla_ESC();  /* PARADA */

       /* Calcular a posio de chegada */
       xc=xp+translac[ind]*cos(teta*((double)(PI/180.0)))*espacx;
       yc=yp-translac[ind]*sin(teta*((double)(PI/180.0)))*espacx;
	  /* Mostrar o trajeto entre os pontos de partida e chegada */
	  setcolor(WHITE); /* cor branca */
	  line(xp,yp,xc,yc);
	  /* Desenhar a seta mostrando o ngulo no ponto de chegada */
	  desenhar_seta(xp,yp,teta,cor[2]);
	  desenhar_seta(xc,yc,teta,cor[2]);
	  colocar_texto(2,YELLOW); /* escrever "translao" */
       /* acionar hardware da plataforma para fazer movimento de rotao */
       acionar_hard_plat(0,translac[ind]);
	  testar_tecla_ESC();

       /* Preparar para prximo ponto da trajetria */
       xp=xc; /* Fazer ponto de partida = ponto de chegada */
       yp=yc;
	  colocar_texto(1,BLACK);  /* apagar texto "rotao"   */
	  colocar_texto(2,BLACK);  /* apagar texto "translao */
       }

   /* fechar arquivo aberto para escrita da tabela de movimentos*/
   if (CodEscrita == 1) fclose(foutptr);

   preparar_topo_tg(FONT_COLOR);
   outtext("FIM ... Pressione ESC para finalizar ...");
   setcolor(WHITE);
   settextjustify(RIGHT_TEXT,TOP_TEXT);
   moveto(mx,dy+(nl-1)*espacy);
   outtext("ESC");
   while (getch() != ESC);

} /* FIM DA ROTINA */
/*-------------------------------------------------------------------------*/
/* Rotina para inicializar o hardware da plataforma.
   Assume-se que o usurio  responsvel por verificar a presena da placa
   de acionamento no micro.
   Para a plataforma ser acionada o programa deve ser executado
   tendo como nico argumento a palavra "acionar_plataforma"
   (ex: romeo acionar_plataforma)
   */
void inicializar_hard_plat()
{
   int value;
   /* comando para programao do chip 8255A (vide TG Eiji/97, pag. 23):
   endereo: A11-A0 = 771 (base 10) = 0011 0000 0011 (base 2)
   dados:     D7-D0 = 128 (base 10) =      0100 0000 (base 2) */
   if (strcmpi(cmd_ac_plat,AC_PLAT)==0)
      {
      value=outp(771,128);
      /* testar erro na inicializao. Melhorar este teste de
	 erro pois pode ser que value=128 mesmo se ocorreu erro
	 na execuo da instruo outp */
      if (value!=128)
	 {
	 INI_PLAT_ERROR=1;
	 preparar_topo_tg(FONT_COLOR_ERROR);
	 outtext("Erro na inicializao da plataforma !!!");
	 getch();
	 preparar_topo_tg(FONT_COLOR);
	 outtext("A plataforma no ser acionada. Pressione ESC para continuar ...");
	 }
      else
	 INI_PLAT_ERROR=0;
      }

} /* FIM DA ROTINA */
/*-------------------------------------------------------------------------*/
/* Rotina para acionar o hardware da plataforma.
   Primeiro executa-se o movimento de rotao e depois o movimento de
   translao. Como ponto de referncia adota-se que o ponto de vista
   do observador est em cima da plataforma olhando para baixo.
   Entradas:
     rot_gr: rotao a ser executada, medida em graus.
	     rot_gr > 0: rotao no sentido anti-horrio.
	     rot_gr < 0: rotao no sentido horrio.
     transl: translao a ser executada, medida em unidades da distncia.
	     horizontal do labirinto.
	     transl > 0: translao para frente.
	     transl < 0: translao em r.
   Assume-se que o usurio  responsvel por verificar a presena da placa
   de acionamento no micro.
   Para a plataforma ser acionada o programa deve ser executado
   tendo como nico argumento a palavra "acionar_plataforma"
   (ex: romeo acionar_plataforma)
   */
void acionar_hard_plat(float rot_gr,float transl)
{
   /* comando para programao do chip 8255A (vide TG Eiji/97, pag. 24):
   endereo: A11-A0 = 768 (base 10) = 0011 0000 0000 (base 2)
   dados:     D7-D4 =   0 (base 10) = 0000 (base 2)
	      D3 e D2 acionam motor da direita.
	      D1 e D0 acionam motor da esuqerda.
	      D3 e D1 determinam o tipo de movimento. Qualquer movimento
	       executado "subindo e descendo" os valores de D2 e D0.
	      Se D3 EQ D1: --> translao:
			   Se D3 = 1, frente;
			   Se D3 = 0; r;
	      Se D3 DIF D1 --> rotao:
			   Se D3 = 1, giro para a direita  (sentido horrio);
			   Se D3 = 0, giro para a esquerda (sentido anti-horrio);
      Frente: D3-D1 =  10 (base 10) = 1010 (base 2), D3=D1=1, D2 e D0 mudam
		    =  15 (base 10) = 1111 (base 2)
	  R: D3-D1 =   0 (base 10) = 0000 (base 2), D3=D1=0, D2 e D0 mudam
		    =   5 (base 10) = 0101 (base 2)
Anti-Hor (+): D3-D1 =   2 (base 10) = 0010 (base 2), D3=0, D1=1, D2 e D0 mudam
  (Esquerda)	    =   7 (base 10) = 0111 (base 2)
 Horrio (-): D3-D1 =   8 (base 10) = 1000 (base 2), D3=1, D1=0, D2 e D0 mudam
  (Direita)	    =  13 (base 10) = 1101 (base 2)
   */
   int NP,i,c1,c2,value;

   if (strcmpi(cmd_ac_plat,AC_PLAT)==0 && INI_PLAT_ERROR==0)
      {

      /* MOVIMENTO DE ROTAO */
      NP=(rot_gr/360.0)*N_passos_rot_360;
      NP=abs(NP);
      if (rot_gr>0) {c1=2; c2= 7;} /* rotao +, AH, para a esquerda (2, 7) */
      if (rot_gr<0) {c1=8; c2=13;} /* rotao -,  H, para a direita  (8,13) */
      for (i=0;i<NP;i++)
	  {
	  value=outp(768,c1);
	  delay(DELAY_PULSO);
	  value=outp(768,c2);
	  delay(DELAY_PULSO);
	  }
      /* FIM MOVIMENTO DE ROTAO */

      /* MOVIMENTO DE TRANSLAO */
      NP=transl*N_passos_desl_horz;
      NP=abs(NP);
      if (transl>0) {c1=10; c2=15;} /* translao +, para frente (10, 15) */
      if (transl<0) {c1= 0; c2= 5;} /* translao -,      de r  ( 0,  5) */
      for (i=0;i<NP;i++)
	  {
	  value=outp(768,c1);
	  delay(DELAY_PULSO);
	  value=outp(768,c2);
	  delay(DELAY_PULSO);
	  }
      /* FIM MOVIMENTO DE TRANSLAO */

      value++; /* instruo no operacional includa apenas para
		  tirar o warning do compilador sobre value */
      } /* fim do if que testa se a plataforma deve ser acionada */

} /* FIM DA ROTINA */
/*-------------------------------------------------------------------------*/
/* Rotina para preparar o topo da tela grfica onde  escrito texto
   com os valores de g,h e g+h para cada n durante a procura da
   rota e o valor final do custo da rota encontrada.
   Esta rotina  chamada pelas rotinas expand() e tracar_trajetoria */
void preparar_topo_tg(int indcor)
{
  int i;
  //char BUF_TEMP[80];

  /* Apagar o que estiver escrito no top da janela grfica */
  setcolor(BLACK); /* cor 0 = black */
  for (i=0;i<dy-raio-1;i++) line(0,i,mx-1,i);
  //setcolor(BLACK);
  //settextstyle(FONT_TYPE,HORIZ_DIR,FONT_SIZE);
  //settextjustify(CENTER_TEXT,TOP_TEXT);
  //moveto(mx/2,(dy-raio)/2);
  //for (i=0;i<79;i++) BUF_TEMP[i]='\xdb';
  //BUF_TEMP[79]='\0';

  /* Definir cor, estilo e justificacao do texto   */
  /* a ser colocado no topo da tela grfica        */
  setcolor(indcor);
  settextstyle(FONT_TYPE,HORIZ_DIR,FONT_SIZE);
  settextjustify(CENTER_TEXT,TOP_TEXT);
  moveto(mx/2,(dy-raio)/2);

} /* FIM DA ROTINA */
/*-----------------------------------------------------------------------*/
/* Rotina para desenhar 1 seta usando linhas
   Esta rotina  chamada pelas rotinas:
      desenhar_labirinto(): para mostrar a orientao inicial TETA0, e
      expand(): para mostrar a orientao sendo considerada durante
		a expanso do n,
      show_path: para mostrar a orientao em cada ponto da trajetria.
   Dados de entrada:
      cx, cy = coordenadas do ponto de onde parte a seta
      tetag = ngulo da seta com o horizontal em graus
      indcor = nmero que indica a cor da seta
*/
void desenhar_seta(int cx,int cy, float tetag, int indcor)
{
  float L1,L2,L3,tetar;
  float P1[2],P2[2],P3[2],P4[2];  /* cordenadas (x,y) que definem */
				  /* os pontos principais da seta */
  /*	        (P3)
		 | \            |\         L1 = comprimento da seta
  (cx,cy)    (P2)|   \       L3 |  \       L2,L3 = catetos que definem
    ------------------>(P1)     |____\             os lados da seta
	  L1     |   /            L2
		 | /
		(P4)

  teta = ngulo da reta que forma a base da seta com a horizontal.
	 teta  positivo no sentido anti-horrio.
  */
  tetar=((float)(PI/180.0))*tetag;
  L1=2.0*((float)(raio));
  L2=0.5*((float)(raio));
  L3=0.3*((float)(raio));

  /* (cx,cy) esto no p da seta, P1 est no bico da seta */
  P1[0]=L1*cos(tetar);       P1[1]=L1*sin(tetar);
  P2[0]=(L1-L2)*cos(tetar);  P2[1]=(L1-L2)*sin(tetar);
  P3[0]=P2[0]-L3*sin(tetar); P3[1]=P2[1]+L3*cos(tetar);
  P4[0]=P2[0]+L3*sin(tetar); P4[1]=P2[1]-L3*cos(tetar);

  /* as coordenadas do eixo y, P1[1], P3[1] e P4[1],
     so invertidas porque o eixo y aponta para baixo na tela
     e o ponto (0,0) est no canto superior esquerdo da tela */

  setcolor(indcor);
  /* definio das coordenadas da base da seta */
  line(cx,cy,cx+P1[0],cy-P1[1]);  /* base da seta */

  line(cx+P1[0],cy-P1[1],cx+P3[0],cy-P3[1]);
  line(cx+P1[0],cy-P1[1],cx+P4[0],cy-P4[1]);

} /* FIM DA ROTINA */
/*-----------------------------------------------------------------------*/
/* Rotina para desenhar um crculo no canto superior direito na
   cor desejada. Neste programa as cores sao:
   verde: para indicar que o caminho sendo considerado  melhor que
	  o caminho armazenado
   vermelho: para indicar que o caminho sendo considerado  pior ou igual
	  ao caminho armazenado e ser descartado
   branco: se o caminho for novo.
*/
void desenhar_circ(int indcor)
{
   if ((pap=='s') || (pap=='S'))
      {
      setcolor(indcor);
      circle(dx+(nc-1)*espacx+dx/2,dy,raio/2);
      setfillstyle(SOLID_FILL,indcor);
      floodfill(dx+(nc-1)*espacx+dx/2,dy,indcor);
      }
}
/*-----------------------------------------------------------------------*/
/* Rotina que, no caso de usurio pedir expanso passo a passo, verifica
   se a tecla apertada pelo usurio foi a tecla ESC. Neste caso o programa
    abortado e uma msg  impressa na tela de texto nornal. */
void testar_tecla_ESC()
{
    int x,y;
    /* Imprime ESC canto direito inferior */
    setcolor(WHITE);
    settextjustify(RIGHT_TEXT,TOP_TEXT);
    x=mx; y=dy+(nl-1)*espacy;
    moveto(x,y);
    outtext("ESC");
   /* Verifica se a tecla ESC foi apertada */
   if (getch() == ESC)
      {
      closegraph();
      beep(); /* BEEP */
      textcolor(YELLOW);
      cprintf("\r\n\r\n Programa interrompido pelo usurio ...\n");
      textcolor(WHITE); normvideo(); /* select normal intensity characteres */
      cprintf("\r\n");  /* imprime estes caracteres para voltar o texto
			   ao modo normal (branco com intensidade normal) */
      exit(1);
      }
   else /* apagar a palavra ESC */
      {
      setcolor(BLACK);
      moveto(x,y);
      outtext("ESC");
      }
}
/*-----------------------------------------------------------------------*/
/* Rotina para colocar uma palavra na posio cx,cy da tela grfica
   com a cor desejada. Neste caso  usado o seguinte cdigo:
   se cod = 1, coloca a palavra "rotao" no canto superior direito
   se cod = 2, coloca a palavra "translao" no canto superior direito
	       mas logo abaixo da palavra rotao.
   Para apagar as palavras, basta reescrev-las na cor preta.
*/
void colocar_texto(int cod, int cor)
{
    setcolor(cor);
    settextjustify(RIGHT_TEXT,TOP_TEXT);
    if (cod==1)
       {
       moveto(mx,dy);
       outtext("rotao");
       }
    else if (cod==2)
       {
       moveto(mx,dy+2*textheight("rotao"));
       outtext("translao");
       }
    else
       beep(); /* BEEP */
}
/*-------------------------------------------------------------------------*/
void beep()
{
/* #if #elif e #endif so comandos de compilao condicional para
   evitar os warnings to TURBO C compiler */
#if BEEP_CONST==1
     printf("\x7");
#elif BEEP_CONST==2
     sound(BEEP_FREQ);
     delay(BEEP_DUR);
     nosound();
#endif

} /* end rotina beep */
/*-------------------------------------------------------------------------*/

/* ----> FIM DO PROGRAMA <---- */

