% MATLAB script 5.x - file: CODER.M
% Treinamento do problema "Grandmother Coding" na entrada e na saida da rede neural
% Usando uma camada escondida com menos unidades que na camada de entrada,
% a camada escondida e' forcada a desenvolver um codigo de compressao.
%
% Metodo de treinamento: BP
% Programa valido apenas para 1 camada escondida
%   sem ligacoes diretas entre as camadas de entrada e de saida
% Assumindo que unidades escondidas e de saida usam funcao tanh(x)
%
% Rotinas auxiliares: bpc.m, calcer.m, nnout.m, ord.m
% Notacao: MATRIZES: todas as letras maiusculas
%          VetoreS: todos sao vetores coluna
%                   apenas letras inicial e/ou final sao maiusculas
%          elementos: todas as letras minusculas
%
% Autor: Cairo L. Nascimento Jr. - ITA - cairo@ita.br
% Data: 11/Out/2001, Versao: 1.0 - Matlab 5.x

% Preparacao para posicionamento das figuras
clear; close all;
figure(1); set(gcf,'Position',[  7   245   390   283]);
set(gcf,'DoubleBuffer','on'); % habilitar tracado mais rapido das linhas
                              % para mostrar animacoes em maquinas mais rapidas
figure(2); set(gcf,'Position',[403   245   390   283]);
set(gcf,'DoubleBuffer','on');
figure(3); set(gcf,'Position',[  7    33   390   136]);
set(gcf,'DoubleBuffer','on');
figure(4); set(gcf,'Position',[403    33   390   136]);
set(gcf,'DoubleBuffer','on');

% Descricao das variaveis globais
% Estas variaveis globais sao usadas nas rotinas auxliares
global iter npat nhid nout netah netao ninp netadec biasder nord;

% Matriz INPD = entrada dos padroes a serem aprendidos
% Matriz OUTD = saida desejada para os padroes a serem aprendidos
% Cada padrao e' uma coluna
% numero de padroes = numero de colunas
% numero de unidades de entrada = numero de linhas
N=8;
INPD=2*eye([N,N])-ones([N,N]);
OUTD=INPD;
if N == 8; nhid=3; end;      % numero de unidades escondidas

% Definicao dos parametros para o algoritmo de treinamento
netah=0.4/2;  % learning rates para as 2 camadas de pesos
netao=0.8/2;
erromax=0.02; % maximo erro rms permitido para parar o treinamento da rede neural
niter=600;    % numero maximo de iteracoes de treinamento
nord=1;       % nord = 0, > 0 : ordenacao sequencial ou aleatoria
netadec=0;    % decay rate
biasder=0.1;  % bias a ser adicionado na derivada da funcao de ativacao
ndisp=10;     % periodo de atualizacao da saida grafica
Nlevels=256;  % no. de niveis usados para a visualizacao grafica
              % Nlevels > 1

% ninp = numero de unidades de entrada
% nout = numero de unidades de saida
% nhid = numero de unidades da camada escondida
% npat = numero de padroes a serem aprendidos pela rede
% nunits = numero total de unidades
[ninp,npat]=size(INPD);
[nout,npat]=size(OUTD);
nunits = ninp+nhid+nout;

% Matriz WHI = matriz de pesos entre as camadas de entrada e escondida
% Matriz WOH = matriz de pesos entre as camadas escondida  e de saida
% BiasH = vetores coluna com os bias da camada escondida
% BiasO = vetores coluna com os bias da camada de saida
% Notacao: W(i,j) = peso da unidade i vindo da unidade j
% Os pesos recebidos pela unidade i estao contidos na linha  i
% Os pesos enviados  pela unidade j estao contidos na coluna j
% Geracao da rede inicial = matrizes de pesos e vetores de bias
rand('state',sum(100*clock));
%rand('state',2);
WHI=2*rand(nhid,ninp)-1;
WOH=2*rand(nout,nhid)-1;
BiasH=2*rand(nhid,1)-1;
BiasO=2*rand(nout,1)-1;

% Definicao de variaveis auxiliares
  bn=0;   % bn significa "break now"

  % Labels for the axes of figures 2, 3 and 4: YL2, XL3 and YL3
  for i=1:nhid; YL2(i)={['H' num2str(i)]}; end;  % Y label for figure(2)
  for i=nhid+1:nhid+N; YL2(i)={['O' num2str(i-nhid)]}; end;
  for i=1:N; XL3(i)={['I' num2str(i)]}; end;     % X label for figure(3)
  XL3(N+1)={['BiasH']};
  for i=1:nhid; YL3(i)={['H' num2str(i)]}; end;  % Y label for figure(3)
  XL4=YL3;                                       % X label for figure(3)
  XL4(nhid+1)={['BiasO']};
  for i=1:N; YL4(i)={['O' num2str(i)]}; end;

  % Colormap for figure(2): -1 (white) : 1 (black)
  map2=1-gray(Nlevels);  % size(map2) = [Nlevels,3]
  % Colormap for figures 3 and 4:
  % if x=max(abs(W)), then colormap is: -x(red) : 0(white) : x(green)
  Nl2=fix(Nlevels/2);  % Nlevels > 1, Nl2 >= 1
  map3(1:Nl2,:)=gray(Nl2); map3(1:Nl2,1)=ones([Nl2 1]); % red scale (deep red to white)
  map3(Nl2:2*Nl2-1,:)=1-gray(Nl2); map3(Nl2:2*Nl2-1,2)=ones([Nl2 1]); % reverse green scale (white to deep green)
  % size(map3) = [2*Nl2-1,3]

% Loop das iteracoes de treinamento
for iter=0:niter;

	   % Calculo do erro nesta iteracao
	     [errop,OutpatO,OutpatH]=calcer(WHI,WOH,BiasH,BiasO,INPD,OUTD);
        erro(iter+1,1)=errop;
        if (errop <= erromax) | (iter >= niter); bn=1; end;
        
      % Saida grafica durante o treinamento
        if ndisp*fix(iter/ndisp)==iter | iter == 0 | bn == 1;
           % Plote a curva de aprendizado
           figure(2);
           semilogy([0:length(erro)-1],erro); hold on;
           XLim=get(gca,'XLim'); YLim=get(gca,'YLim');
           plot(0,erro(1),'*');
           plot([0 XLim(2)],[erromax erromax],'r-','LineWidth',2);
           hold off;
           set(gca,'YLim',[erromax/2 YLim(2)]); YLim=get(gca,'YLim');
           if YLim(2) >= 10*max(erro); set(gca,'YLim',[YLim(1) YLim(2)/10]); end;
           xlabel('No. da iterao');
           ylabel('Erro Quadrtico Mdio');
           title(['Curva de Aprendizado, Erro = ' num2str(errop)]);
           grid;
           drawnow;
           % Mostrar as sadas das unidades nas camadas escondidas
           % e de saida da rede neural
           figure(1); % white = -1, black = 1
           maxy=1; miny=-1;
           colormap(map3);
           %OutImage=[OutpatH;OutpatO]*(Nlevels-1)/(maxy-miny);
           %OutImage=OutImage+Nlevels-(Nlevels-1)*maxy/(maxy-miny);
           OutImage=([OutpatH;OutpatO]-miny)/(maxy-miny);
           OutImage=OutImage*(Nlevels-1)+1;
           image(OutImage);
           set(gca,'XTick',[1:N]);
           set(gca,'XTickLabel',[1:N]);
           set(gca,'YTick',[1:(N+nhid)]);
           set(gca,'YTickLabel',YL2);
           title('Sadas das camadas escondida e de sada');
           xlabel('No. do padro de entrada');
           ylabel('No. da unidade');
           drawnow;
           % Mostrar os pesos recebidos pelas unidades nas camadas escondidas
           figure(3);
           maxy=max(max(abs([WHI,BiasH]))); miny=-maxy;
           colormap(map3);
           OutImage=([WHI,BiasH]-miny)/(maxy-miny);
           OutImage=OutImage*(2*Nl2-1-1)+1;
           image(OutImage);
           set(gca,'XTick',[1:(N+1)]);
           set(gca,'XTickLabel',XL3);
           set(gca,'YTick',[1:nhid]);
           set(gca,'YTickLabel',YL3);
           title('Pesos entre as camada de entrada e escondida');
           xlabel('No. da unidade de entrada');
           ylabel('No. da unidade escondida');
           drawnow;
           % Mostrar os pesos recebidos pelas unidades na camada de saida
           figure(4);
           maxy=max(max(abs([WOH,BiasO]))); miny=-maxy;
           colormap(map3);
           OutImage=([WOH,BiasO]-miny)/(maxy-miny);
           OutImage=OutImage*(2*Nl2-1-1)+1;
           image(OutImage);
           set(gca,'XTick',[1:(nhid+1)]);
           set(gca,'XTickLabel',XL4);
           set(gca,'YTick',[1:N]);
           set(gca,'YTickLabel',YL4);
           title('Pesos entre as camada escondida e de sada');
           xlabel('No. da unidade escondida');
           ylabel('No. da unidade de sada');
           drawnow;
           %
           if iter==0;
              figure(1); hold on;
              XLim=get(gca,'XLim');
              plot([0 XLim(2)],[erromax erromax],'r-');
              plot(0,erro(1),'r*');
              hold off;
              disp('Pausing ...'); pause(1); disp('...Out of pause');
           end;  
        end;         
      % Treinar a rede neural se necessario
        if bn == 0;
     	     % Chamada da rotina Back Propagation
             [WHI,WOH,BiasH,BiasO]=bpc(WHI,WOH,BiasH,BiasO,INPD,OUTD);
        else;
           break;     
        end;
        
end; % end "for iter=0:niter;"
     
% Mostre curva de aprendizado ao termino do aprendizado
  figure(2);
  title(['Curva de Aprendizado, Erro = ' num2str(errop), ', Niter = ' num2str(iter)]);
  drawnow;    

% EOF CODER.M
