function [B,C]=flip(A,probflip)
% Funcao usada para adicionar ruido do tipo flip-flop
% em uma matriz.
% Dada uma matriz A, a matriz B e' formada pelos elementos
% da matriz A onde cada elemento de A teve o sinal invertido
% com probabilidade dada pela variavel probflip com distribuicao 
% uniforme, onde 0 <= probflip <= 1.
% A matriz C indica quais elementos de A forma modificados.
% Se C(i,j)=1, A(i,j) foi modificado, caso contrario C(i,j)=0
% Autor: Cairo L. Nascimento Jr. - ITA - 14/Jun/1995
% Versao: 1.0

if probflip < 0 | probflip > 1;
   error('probflip fora do intervalo [0,1].');
end;

% gera matriz com numeros aleatorios uniformente
% distribuidos entre 0 e 1.
B=rand(size(A));
% B<probflip = matriz com 0 e 1 onde 1 significa que o
% elemento de B e' menor que probflip, assim este
% elemento de A deve ter o sinal invertido.
% Assim se C=(B<probflip), se C(i,j)=0, B(i,j)=A(i,j);
% se C(i,j)=1; B(i,j)=-A(i,j).
C=(B<probflip);
B=A.*(1-2*C);

% EOF flip.m
