% MATLAB script 5.x - file: XOR5.M
% Treinamento do XOR
% Metodo de treinamento: BP
% Programa valido apenas para 1 camada escondida
%   sem ligacoes diretas entre as camadas de entrada e de saida
% Assumindo que unidades escondidas e de saida usam funcao tanh(x)
%
% Rotinas auxiliares: bpc.m, calcer.m, nnout.m, ord.m
% Notacao: MATRIZES: todas as letras maiusculas
%          VetoreS: todos sao vetores coluna
%                   apenas letras inicial e/ou final sao maiusculas
%          elementos: todas as letras minusculas
%
% Autores: Eduardo Botter e Cairo L. Nascimento Jr. - ITA - cairo@ita.br
% Data:  7/Set/1995, Versao: 1.0 - Matlab 4.2
% Data: 14/Out/1998, Versao: 1.1 - Matlab 4.2
% Data:  2/Jul/2001, Versao: 1.2 - Matlab 5.3
% Data: 17/Fev/2002, Versao: 1.3 - Revisao

% Preparacao para posicionamento das figuras
clear all; close all; Nf=1;
figure(Nf);   set(gcf,'Position',[  5  305   395   225]);
set(gcf,'DoubleBuffer','on'); % habilitar tracado mais rapido das linhas
                              % para mostrar animacoes em maquinas mais rapidas
figure(Nf+1); set(gcf,'Position',[405  305   395   225]);
set(gcf,'DoubleBuffer','on');
figure(Nf+2); set(gcf,'Position',[  5   35   395   195]);
set(gcf,'DoubleBuffer','on');
figure(Nf+3); set(gcf,'Position',[ 405  35   395   195]);
set(gcf,'DoubleBuffer','on');

% Matriz INPD = entrada dos padroes a serem aprendidos
% Matriz OUTD = saida desejada para os padroes a serem aprendidos
% Cada padrao e' uma coluna
% numero de padroes = numero de colunas
% numero de unidades de entrada = numero de linhas

INPD=[-1  1 -1  1
      -1 -1  1  1];
OUTD=[-1  1  1 -1];

% Descricao das variaveis globais
% Estas variaveis globais sao usadas nas rotinas auxliares
global iter npat nhid nout netah netao ninp netadec biasder nord;

nhid=2;       % numero de unidades escondidas
netah=0.4/2;  % learning rates para as 2 camadas de pesos
netao=0.8/2;
erromax=0.1;  % maximo erro rms permitido para parar o treinamento da rede neural
niter=300;    % numero maximo de iteracoes de treinamento
nord=1;       % nord = 0, > 0 : ordenacao sequencial ou aleatoria
netadec=0;    % decay rate
biasder=0.1;  % bias a ser adicionado na derivada da funcao de ativacao

% ninp = numero de unidades de entrada
% nout = numero de unidades de saida
% nhid = numero de unidades da camada escondida
% npat = numero de padroes a serem aprendidos pela rede
% nunits = numero total de unidades
[ninp,npat]=size(INPD);
[nout,npat]=size(OUTD);
nunits = ninp+nhid+nout;

% Matriz WHI = matriz de pesos entre as camadas de entrada e escondida
% Matriz WOH = matriz de pesos entre as camadas escondida  e de saida
% BiasH = vetores coluna com os bias da camada escondida
% BiasO = vetores coluna com os bias da camada de saida
% Notacao: W(i,j) = peso da unidade i vindo da unidade j
% Os pesos recebidos pela unidade i estao contidos na linha  i
% Os pesos enviados  pela unidade j estao contidos na coluna j
% Geracao da rede inicial = matrizes de pesos e vetores de bias
rand('state',sum(100*clock));
%rand('state',2);
WHI=2*rand(nhid,ninp)-1;
WOH=2*rand(nout,nhid)-1;
BiasH=2*rand(nhid,1)-1;
BiasO=2*rand(nout,1)-1;

% Loop das iteracoes de treinamento
ndisp=5;   % periodo de atualizacao da saida grafica
iter=0;

while iter <= niter;

	   % Calculo do erro nesta iteracao
      [errop,OutpatE,OutpatH]=calcer(WHI,WOH,BiasH,BiasO,INPD,OUTD);      
      erro(iter+1,1)=errop;
        
      % Mostrar cabecalho adequado na janela de comandos se 1a iteracao
        if iter==0;
           OutpatE0=OutpatE;
           disp('Iterao: Sada da rede e Erro RMS');
        end;

      % Visualizacao grafica das saidas durante a aprendizagem
        if ndisp*fix(iter/ndisp)==iter | iter == 0 | errop < erromax;
           fprintf('Iter = %4i, Sada: %6.3f %6.3f %6.3f %6.3f, RMS Erro: %g \n', ...
                    iter,OutpatE(1,1), OutpatE(1,2), OutpatE(1,3), OutpatE(1,4), erro(iter+1,1));
           %
           figure(Nf);
           s='rkb';
           for ipat=1:npat;
               for ip=1:ninp;
                   d=(ip-1)/10;
                   hp(ip)=plot([ipat+d ipat+d],[0 INPD(ip,ipat)],s(ip));
                   hold on;
                end;
                hp(ip+1)=plot([ipat+0.2 ipat+0.2],[0 OutpatE(:,ipat)],s(ip+1));
                set(hp,'LineWidth',2);
           end;
           legend(hp,'I1','I2','O1',0);
           plot([0 npat+1],[0 0]); hold off;
           axis([0 npat+1 -1 1]);
           set(gca,'XTick',[0:npat+1],'XTickLabel',[' 1234 ']');
           %title(['Iterao = ' num2str(iter) ', Erro = ' num2str(errop)]);
           xlabel('No. do padro');
           title('Entradas e Sada da Rede Neural');
           grid;
           drawnow;
           %
           figure(Nf+3);
           xds=[-2; 2];
           maxow=max(abs([WOH BiasO]));
           for ih=1:nhid+nout; % plot decision surfaces for the 2 hidden and 1 output units
               if ih <= nhid; wds=WHI(ih,:); biasds=BiasH(ih); end;
               if ih > nhid;  wds=WOH(ih-nhid,:); biasds=BiasO(ih-nhid); end;                              
               yds(1)=-(wds(1)*xds(1)+biasds)/wds(2);
               yds(2)=-(wds(1)*xds(2)+biasds)/wds(2);                
               plot(xds,yds,s(ih)); hold on;
               % show the positive and negative sides of the decision surface
               for xd=xds(1):0.3:xds(2);
                   clear ht;
                   yd=-(wds(1)*xd+biasds)/wds(2);
                   yh=tanh(wds(1)*xd+wds(2)*(yd+0.15)+biasds);
                   if yh > 0 & abs(yd+0.15) < 2;
                      ht=text(xd,yd+0.15,'+','Color',s(ih));
                   end;
                   if yh < 0 & abs(yd-0.15) < 2;
                      ht=text(xd,yd-0.15,'+','Color',s(ih));   
                   end;
               end;
               % plot output weight for this hidden unit
               if ih <= nhid;
                  d=1+0.25*ih;
                  hw(ih)=plot([d d],[1 1+WOH(ih)/maxow],s(ih));
               end;
           end;
           % plot output bias
           hw(nhid+1)=plot([1.75 1.75],[1 1+BiasO/maxow],s(nhid+1));
           %set(hw,'LineWidth',2);
           text(2,2,'O1','Color',s(nhid+1),'HorizontalAlignment','right', ...
                'VerticalAlignment','top');
           for ih=1:nhid; % plot hidden unit weights as vertical lines
               WHI2=[WHI(ih,:) BiasH(ih)]; % hidden unit weight & hidden unit bias
               maxhw=max(abs(WHI2));
               d=3*ih-5;
               for ip=1:ninp+1;
                   d1=d+ip/(ninp+2);
                   hw(ip)=plot([d1 d1],[-1 -1+WHI2(ip)/maxhw],s(ip));
               end;
               %set(hw,'LineWidth',2);
               ht(ih)=text(d,0,['H' num2str(ih)],'Color',s(ih),'VerticalAlignment','top');
               if ih==2*fix(ih/2);
                  set(ht(ih),'Position',[d+1 0],'HorizontalAlignment','right');
               end;
           end;
           % marque os pontos de treinamento na entrada
           for ipat=1:npat; 
               if OUTD(ipat) > 0; st='rs'; else; st='ko'; end;
               plot(INPD(1,ipat),INPD(2,ipat),st,'MarkerFaceColor',st(1));
           end;
           % marque como os pontos de treinamento na entrada sao remapeados
           % pelas unidades escondidas
           % para que o texto fica acima da marcas os loops acima e abaixo
           % devem ser 1 apos o outro e nao devem ser combinados em 1 so'
           for ipat=1:npat;
               text(OutpatH(1,ipat),OutpatH(2,ipat),num2str(ipat), ...
                    'HorizontalAlignment','center','Color',s(3), ...
                    'FontWeight','bold');
           end;
           set(gca,'XTick',[-1 1],'YTick',[-1 1]);
           set(gca,'Xlim',[-2 2],'Ylim',[-2 2]);            
           xlabel('I1/H1');
           ylabel('I2/H2');
           title('Superfcies de Deciso e Pesos das Unidades');
           grid;
           hold off;
           drawnow;
           %
           figure(Nf+2); % plot output of the hidden units
           for ipat=1:npat;
               for ih=1:nhid;
                   d=(ih-1)/10;
                   hp(ih)=plot([ipat+d ipat+d],[0 OutpatH(ih,ipat)],s(ih)); hold on;
               end;
               set(hp,'LineWidth',2);
           end;
           plot([0 5],[0 0]);  % horizontal line           
           hold off;
           axis([0 5 -1 1]);
           set(gca,'XTick',[0:5],'XTickLabel',[' 1234 ']');
           legend(hp,'H1','H2',0);
           xlabel('No. do padro');
           title('Sadas e Pesos das Unidades da Camada Escondida');
           grid;
           drawnow;
           %
           figure(Nf+1);
           semilogy([0:length(erro)-1],erro); hold on;
           XLim=get(gca,'XLim'); YLim=get(gca,'YLim');
           plot(0,erro(1),'*');
           plot([0 XLim(2)],[erromax erromax],'r-','LineWidth',2);
           hold off;
           set(gca,'YLim',[erromax/2 YLim(2)]); YLim=get(gca,'YLim');
           if YLim(2) >= 10*max(erro); set(gca,'YLim',[YLim(1) YLim(2)/10]); end;
           xlabel('No. da iterao');
           ylabel('Erro Quadrtico Mdio');
           title(['Curva de Aprendizado, Erro = ' num2str(errop)]);
           grid;
           drawnow;
           if iter == 0; % Pause para mostrar solucao antes do aprendizado
              disp('Pausing ...'); pause(2); disp('...Out of pause');
           end; % end "if iter == 0;"
           end; % end "ndisp*fix(iter/ndisp)==iter | iter == 0;" 
         % Treinar a rede neural se necessario
         if errop > erromax;
     	      % Chamada da rotina Back Propagation
              [WHI,WOH,BiasH,BiasO]=bpc(WHI,WOH,BiasH,BiasO,INPD,OUTD);
         else;
            break;
         end;
         
         iter=iter+1;
         
end; % end "iter <= niter;"

% Mostre curva de aprendizado ao termino do aprendizado
  figure(Nf+1);
  title(['Curva de Aprendizado, Erro = ' num2str(errop), ', Niter = ' num2str(iter)]);
  drawnow;

% Mostre a saida inicial e final e a saida desejada;
  disp(' ');
  disp('Sada Inicial, Sada Final, Sada desejada');
  [OutpatE0; OutpatE; OUTD]

% EOF XOR5.M
