% MATLAB script 5.x - file: XOR5B.M
% Treinamento do XOR
% Metodo de treinamento: BP
% Programa valido apenas para 1 camada escondida
%   COM ligacoes diretas entre as camadas de entrada e de saida
% Assumindo que unidades escondidas e de saida usam funcao tanh(x)
%
% Rotinas auxiliares: bpc2.m, calcer2.m, nnout2.m, ord.m
% Notacao: MATRIZES: todas as letras maiusculas
%          VetoreS: todos sao vetores coluna
%                   apenas letras inicial e/ou final sao maiusculas
%          elementos: todas as letras minusculas
%
% Autores: Eduardo Botter e Cairo L. Nascimento Jr. - ITA - cairo@ita.br
% Data: 11/Out/2001, Versao 1.0 - Matlab 5.x
%       17/Fev/2002, Revisao

% Preparacao para posicionamento das figuras
clear all; close all; 
figure(1);set(gcf,'Position',[  5  305   395   225]);
set(gcf,'DoubleBuffer','on'); % habilitar tracado mais rapido das linhas
                              % para mostrar animacoes em maquinas mais rapidas
figure(2);set(gcf,'Position',[405  305   395   225]);
set(gcf,'DoubleBuffer','on');
figure(3);set(gcf,'Position',[  5   35   395   195]);
set(gcf,'DoubleBuffer','on');
figure(4);set(gcf,'Position',[ 405  35   395   195]);
set(gcf,'DoubleBuffer','on');

% Matriz INPD = entrada dos padroes a serem aprendidos
% Matriz OUTD = saida desejada para os padroes a serem aprendidos
% Cada padrao e' uma coluna
% numero de padroes = numero de colunas
% numero de unidades de entrada = numero de linhas

INPD=[-1  1 -1  1
      -1 -1  1  1];
OUTD=[-1  1  1 -1];

% Descricao das variaveis globais
% Estas variaveis globais sao usadas nas rotinas auxliares
global iter npat nhid nout netah netao ninp netadec biasder nord;

nhid=1;       % numero de unidades escondidas
netah=0.8/2;  % learning rates para as 2 camadas de pesos
netao=0.8/2;
erromax=0.1;  % maximo erro rms permitido para parar o treinamento da rede neural
niter=300;    % numero maximo de iteracoes de treinamento
nord=1;       % nord = 0, > 0 : ordenacao sequencial ou aleatoria
netadec=0;    % decay rate
biasder=0.1;  % bias a ser adicionado na derivada da funcao de ativacao

% ninp = numero de unidades de entrada
% nout = numero de unidades de saida
% nhid = numero de unidades da camada escondida
% npat = numero de padroes a serem aprendidos pela rede
% nunits = numero total de unidades
[ninp,npat]=size(INPD);
[nout,npat]=size(OUTD);
nunits = ninp+nhid+nout;

% Matriz WHI = matriz de pesos entre as camadas de entrada e escondida
% Matriz WOI = matriz de pesos entre as camadas de entrada e de saida
% Matriz WOH = matriz de pesos entre as camadas escondida  e de saida
% BiasH = vetores coluna com os bias da camada escondida
% BiasO = vetores coluna com os bias da camada de saida
% Notacao: W(i,j) = peso da unidade i vindo da unidade j
% Os pesos recebidos pela unidade i estao contidos na linha  i
% Os pesos enviados  pela unidade j estao contidos na coluna j
% Geracao da rede inicial = matrizes de pesos e vetores de bias
rand('state',sum(100*clock));
%rand('state',2);
WHI=2*rand(nhid,ninp)-1;
WOI=2*rand(nout,ninp)-1;
WOH=2*rand(nout,nhid)-1;
BiasH=2*rand(nhid,1)-1;
BiasO=2*rand(nout,1)-1;

% Loop das iteracoes de treinamento
ndisp=5;   % periodo de atualizacao da saida grafica
iter=0;

while iter <= niter;

	   % Calculo do erro nesta iteracao
        [errop,OutpatE,OutpatH]=calcer2(WHI,WOI,WOH,BiasH,BiasO,INPD,OUTD);
        erro(iter+1,1)=errop;
        
      % Mostrar cabecalho adequado na janela de comandos se 1a iteracao
        if iter==0;
           OutpatE0=OutpatE;
           disp('Iterao: Sada da rede e Erro RMS');
        end;

      % Visualizacao grafica das saidas durante a aprendizagem
        if ndisp*fix(iter/ndisp)==iter | iter == 0 | errop < erromax;
           fprintf('Iter = %4i, Sada: %6.3f %6.3f %6.3f %6.3f, RMS Erro: %g \n', ...
                    iter,OutpatE(1,1), OutpatE(1,2), OutpatE(1,3), OutpatE(1,4), erro(iter+1,1));
           %
           figure(1);
           s='rkmb';
           for ipat=1:npat;
               for ip=1:ninp;
                   d=(ip-1)/10;
                   hp(ip)=plot([ipat+d ipat+d],[0 INPD(ip,ipat)],s(ip));
                   hold on;
                end;
                hp(ip+1)=plot([ipat+d+0.1 ipat+d+0.1],[0 OutpatH(:,ipat)],s(ip+1));
                hp(ip+2)=plot([ipat+d+0.2 ipat+d+0.2],[0 OutpatE(:,ipat)],s(ip+2));
                set(hp,'LineWidth',2);
           end;
           legend(hp,'I1','I2','H1','O1',0);
           plot([0 npat+1],[0 0]); hold off;
           axis([0 npat+1 -1 1]);
           set(gca,'XTick',[0:npat+1],'XTickLabel',[' 1234 ']');
           %title(['Iterao = ' num2str(iter) ', Erro = ' num2str(errop)]);
           xlabel('No. do padro');
           title('Entradas e Sadas das Unidades da Rede Neural');
           grid;
           drawnow;
           %
           figure(3);
           xds=[-2; 2];
           for ih=1:nhid; % plot decision surfaces
               wds=WHI(ih,:);
               biasds=BiasH(ih);
               yds(1)=-(wds(1)*xds(1)+biasds)/wds(2);
               yds(2)=-(wds(1)*xds(2)+biasds)/wds(2);                
               plot(xds,yds,s(ih)); hold on;
               % show the positive and negative sides of the decision surface
               for xd=xds(1):0.3:xds(2);
                   clear ht;
                   yd=-(wds(1)*xd+biasds)/wds(2);
                   yh=tanh(wds(1)*xd+wds(2)*(yd+0.15)+biasds);
                   if yh > 0 & abs(yd+0.15) < 2;
                      ht=text(xd,yd+0.15,'+','Color',s(ih));
                   end;
                   if yh < 0 & abs(yd-0.15) < 2;
                      ht=text(xd,yd-0.15,'+','Color',s(ih));   
                   end;
               end;
               % show the hidden unit weights as vertical lines
               WHI2=[WHI(ih,:) BiasH(ih)]; % hidden unit weight & hidden unit bias
               maxhw=max(abs(WHI2));
               d=3*ih-5;
               s2='rkb';
               for ip=1:ninp+1;
                   d1=d+ip/(ninp+2);
                   hw(ip)=plot([d1 d1],[-1 -1+WHI2(ip)/maxhw],s2(ip));
               end;
               %set(hw,'LineWidth',2);
               ht(ih)=text(d,0,['H' num2str(ih)],'Color','k','VerticalAlignment','top');
               if ih==2*fix(ih/2);
                  set(ht(ih),'Position',[d+1 0],'HorizontalAlignment','right');
               end;
           end; % end "for ih=1:nhid;"
           % show the weights and bias of the output unit
           WOI2=[WOI(1,:) WOH(1,:) BiasO];
           maxoi=max(abs(WOI2));
           for ip=1:ninp+nhid+1;
               d1=1+ip/(ninp+nhid+2);
               hw(ip)=plot([d1 d1],[-1 -1+WOI2(ip)/maxoi],s(ip));
           end;
           text(2,0,'O1','Color','k','VerticalAlignment','top','HorizontalAlignment','right');
           % marque os pontos de treinamento 
           for ipat=1:npat;                
               if OUTD(ipat) > 0; st='rs'; else; st='ko'; end;
               plot(INPD(1,ipat),INPD(2,ipat),st,'MarkerFaceColor',st(1));
           end;
           set(gca,'XTick',[-1 1],'YTick',[-1 1]);
           set(gca,'Xlim',[-2 2],'Ylim',[-2 2]);            
           xlabel('Entrada I1');
           ylabel('Entrada I2');
           title('Superfcie de Deciso e Pesos da Unidade Escondida');
           grid;
           hold off;
           drawnow;
           %
           figure(4);
           INPD2=[-1  1  1 -1 -1
                  -1 -1  1  1 -1];  % ncols = npat + 1
           OUTD2=[-1  1 -1  1 -1];
           OutpatH2=[OutpatH(1:2),OutpatH(4:-1:3),OutpatH(1)];
           for i=1:npat+1; % calcula a posicao dos 4 pontos do plano
               x(i)=INPD2(1,i);
               y(i)=INPD2(2,i);
               z(i)=WOI(1)*x(i)+WOI(2)*y(i)+BiasO(1);
               z(i)=-z(i)/WOH(1);
           end;
           plot3(x,y,z,'LineWidth',2); hold on; % plota o plano
           ZLim=get(gca,'ZLim'); 
           for i=1:npat; % plota os 4 pontos
               if OUTD2(i) > 0; st='rs'; else; st='ko'; end;
               plot3(x(i),y(i),OutpatH2(i),st,'MarkerFaceColor',st(1));
               plot3([x(i) x(i)],[y(i) y(i)],[ZLim(1) OutpatH2(i)],st(1),'LineWidth',2);
           end;
           hold off; 
           xlabel('I1'); ylabel('I2'); zlabel('H1');
           title('Superfcie de Deciso da Unidade de Sada');
           %set(gca,'ZLim',[-2 2]);
           grid; drawnow; refresh;
           %
           figure(2);
           semilogy([0:length(erro)-1],erro); hold on;
           XLim=get(gca,'XLim'); YLim=get(gca,'YLim');
           plot(0,erro(1),'*');
           plot([0 XLim(2)],[erromax erromax],'r-','LineWidth',2);
           hold off;
           set(gca,'YLim',[erromax/2 YLim(2)]); YLim=get(gca,'YLim');
           if YLim(2) >= 10*max(erro); set(gca,'YLim',[YLim(1) YLim(2)/10]); end;
           xlabel('No. da iterao');
           ylabel('Erro Quadrtico Mdio');
           title(['Curva de Aprendizado, Erro = ' num2str(errop)]);
           grid;
           drawnow;
           if iter == 0; % Pause para mostrar solucao antes do aprendizado
              disp('Pausing ...'); pause(2); disp('...Out of pause');
           end; % end "if iter == 0;"           
         end; % end "ndisp*fix(iter/ndisp)==iter | iter == 0;"
 
         % Treinar a rede neural se necessario
         if errop > erromax;
     	      % Chamada da rotina Back Propagation
              [WHI,WOI,WOH,BiasH,BiasO]=bpc2(WHI,WOI,WOH,BiasH,BiasO,INPD,OUTD);
         else;
            break;
         end;
         
         iter=iter+1;
         
end; % end "iter <= niter;"

% Mostre curva de aprendizado ao termino do aprendizado
  figure(2);
  title(['Curva de Aprendizado, Erro = ' num2str(errop), ', Niter = ' num2str(iter)]);
  drawnow;

% Mostre a saida inicial e final e a saida desejada
  disp(' ');
  disp('Sada Inicial, Sada Final, Sada desejada');
  [OutpatE0; OutpatE; OUTD],

% Mostre os pesos recebidos pela unidade de saida da rede neural
  disp(' ');
  disp('Pesos da rede neural [WHI(2) BiasH(1) WOI(2) WOH(1) BiasO(1)]');
  [WHI BiasH WOI WOH BiasO],

% EOF XOR5B.M
