function [iv,jv]=khout(W,X,N1,N2)
% Funcao para calcular a posicao da unidade vencedora
% de uma rede de Kohonen na grade planar de unidades.
% Assume-se:
%   N1 unidades no eixo horizontal,
%   N2 unidades no eixo vertical,
%   dim W = Ninp,N1*N2,
%   dim X = Ninp,1.
% Obs: No caso das unidades da rede de Kohonen estarem
% arranjadas na forma de linha, faca N2=1.
%
% Entradas:
% W = matriz de pesos da rede de Kohonen.
%    As unidades sao numeradas de baixo para cima, da esquerda
%    para direita. Total de unidades: N1*N2.
%    Coluna k de W corresponde aos pesos recebidos pela unidade
%    de numero k, onde 1 <= k <= N1*N2.
% X = vetor coluna com o padrao de entrada.
%
% Saidas:
% iv,jv = posicao da unidade vencedora na grade N1 x N2.
%
% Autores: Cairo Nascimento (cairo@ita.br) e Eduardo Botter
% Data:  3/Nov/1997 - Verso 1.0 - MATLAB 4.2
%       11/Fev/2002 - Verificao para MATLAB 5.3

[Ninp1,Nunid]=size(W);
[Ninp2,Ncol]=size(X);

if Ninp1 ~= Ninp2 | Nunid ~= N1*N2 | Ncol ~= 1;
   error('KHOUT: Verifique as dimensoes de W e X');
end;

Ninp=Ninp1;

% Similarity Matching: determinar a unidade vencedora
% para esta entrada x, pelo criterio da menor distancia
% [iv,jv] t.q. ||x-w(iv,jv)|| = min(||x-w(i,j)||) p/ todo i e j
uv=1; % definicao da unidade vencedora inicial
dv=sum(abs(X-W(:,uv)));
for i=2:N1*N2;
    d=sum(abs(X-W(:,i)));
    if d < dv; dv=d; uv=i; end;
end;

% Calcular a posicao da unidade vencedora na grade de unidades
% assumindo respectivamente N1 e N2 unidades no eixo horizontal
% e vertical;
iv=ceil(uv/N1);
jv=uv-(iv-1)*N1;

% Testar limites de iv e jv
if iv < 1 | iv > N2 | jv < 1 | jv > N1;
   error('KHOUT: iv e/ou jv fora do intervalo [1,N2],[1,N1]');
end;

% EOF khout.m
