% Programa MATLAB 5.3 - TESTGENK.M
% M-File used to test the generalization of a Kohonen neural network
% Invocar este script depois treinamento da rede pelo script DEMOCRK.M
% Dados de entrada: parametros da rede de Kohonen (W,N1,N2)
%                   padroes de entrada sem ruido (INPD)
%                   UV = vetor com as coordenadas das unidades vencedoras
%                        para entradas sem rudos
%                   niveis de ruido usados para teste (Vprob)
% Dados de saida:   taxas de acerto (TEST(:,4))
% Rotinas auxiliares: FLIP.M, KHOUT.M
% Autor: Cairo L. Nascimento Jr., ITA, cairo@ita.br
% Data: 10/Ago/1995, versao 1.0 para MATLAB 4.2
%        7/Set/1995, versao 1.1 para MATLAB 4.2
%       11/Fev/2002, versao 1.1 para MATLAB 5.3

% Testar se Ncs (numero de clusters sobrepostas) = 0
if Ncs ~= 0;
   error('No. de clusters sobrespostas deve 0 para ser possvel TESTE DE GENERALIZAO !!!');
end;
%
Vprob=-1;
while Vprob < 0 | Vprob > 100;
      % probabilidade de cada pixel mudar de sinal
      Vprob=input('Nvel de rudo em %? ex. [0:5:50]: ');
end;
Vprob=Vprob/100;

if Vprob <= 0 | Vprob > 1;
   Vprob=0;
   ntests=1;
else;
   ntests=0;
   while ntests <= 0;
      % ntests = numero de amostras para cada padrao
      ntests=input('Numero de amostras para cada padro? ex. 50: ');
   end;
end;

limiar=0.45;             % Limiar usado para declarar MISCLASSIFICATION
                         % considerando saidas no intervalo [0,1]

hexastr=['0123456789ABCDEF'];
disp(' ');
disp('Testando a generalizacao da rede neural ...');
ntests0=ntests;

for iprob=1:length(Vprob);
    %rand('seed',iprob); rand('uniform');
    icount=0;     % numero de padroes "misclassified"
    icountOK=0;   % numero de padroes classificados corretamente
    disp(' ');
    disp(['Erros por padrao, Erros totais, Acertos Totais, PFlip = ',num2str(Vprob(iprob)*100),' %']);
    
    for ipat=1:Npat;
        icountp=0;  % contador para apenas este padrao
        ntests=ntests0;
        if Vprob(iprob) <= 0; ntests=1; Vprob(iprob)=0; end;
        if Vprob(iprob) >  1; Vprob(iprob)=1; end;
        [INPDTG,INPDTGC]=flip(INPD(:,ipat)*ones(1,ntests),Vprob(iprob));
        % WARNING: Possivel problema de memoria se ntests for muito grande
        % 2 matrizes de dimensao npat,ntestes
        % Neste caso trocar os loops de posicao ipat e itest e alterar a impressao dos calculos
        for itest=1:ntests;
            Inppat=INPDTG(:,itest);
            % rotina de calculo da saida da rede de Kohonen
            [iv,jv]=khout(W,Inppat,N1,N2);                        
            for iipat=1:Npat;
                d(iipat)=sum(abs(UV(iipat,:)-[iv,jv]));
            end;
            d=d/max(d);
            Outtest=exp(-(N1+N2)*d);
            Out1=(Outtest>(1-limiar)); % componentes de Outtest acima  de 1-limiar
            Out2=(Outtest<limiar);     % componentes de Outtest abaixo de limiar
            Outtest=0.5*(1+Out1-Out2); % se x < limiar -> x=0; se x > 1-limiar; x=1
                                       % se limiar < x < 1-limiar -> x=0.5
            [outmax,imax]=max(Outtest);
            if sum(Outtest==1)==1 & sum(Outtest==0)==(Npat-1) & imax==ipat;
               icountOK=icountOK+1;
            else;
               icount=icount+1; icountp=icountp+1;
            end;
        end;
        fprintf('Teste padrao %s = %g %g %g\n', hexastr(ipat), icountp, icount, icountOK);
    end;
    % 4a coluna de TEST contem as taxas de acerto
    disp('Pflip (%), Erros Totais, Acertos Totais, Taxa de acerto (%)');
    TEST(iprob,:)=[100*Vprob(iprob) icount icountOK 100*icountOK/(icount+icountOK) ];
    fprintf('TEST %g --> %5.3f %6i %6i   %6.4f \n',iprob,TEST(iprob,1:4));
end;

if length(Vprob) > 1;
   figure; clf;
   set(gcf,'Position',[7    34   792   196]);
   set(gcf,'DoubleBuffer','on');
   plot(TEST(:,1),TEST(:,4));
   axis([-inf inf 0 100]); grid; hold on;
   plot(TEST(:,1),TEST(:,4),'*');
   set(gca,'XTick',Vprob*100);
   xlabel('Nvel de Rudo (%)');
   ylabel('Taxa de Acerto (%)');
   title('Teste de Generalizao');
   hold off;
   %print -dmeta acerto;
end;

clear INPDTG INPDTGC % matrizes usada pela rotina FLIP2.M
disp(' ');

% EOF TESTGENK.M
