% Programa MATLAB 5.3 - TPNNK.M
% Permite verificar a saida da rede neural de KOHONEN para um padrao
% definido visualmente e alterado com o mouse.
% Usa os 16 padroes hexadecimais definidos em HEXAPAT.M
% A rede neural de KOHONEN deve ser antes treinada pelo programa DEMOCRK.M
% Para visualizar os 16 padroes hexadecimais (7x7) use o program DSPHEXA2.M
% Para testar a generalizacao da rede neural de KOHONEN use o programa TESTGENK.M
% Autor: Cairo L. Nascimento Jr., cairo@ita.br, ITA
% Data:      11/Fev/2002 - verso 1.0 - MATLAB 5.3

Nf=3;
figure(Nf); clf; set(gcf,'Position',[  5   35   395   195]);
set(gcf,'DoubleBuffer','on');
figure(Nf+1); clf; set(gcf,'Position',[ 405  35   395   195]);
set(gcf,'DoubleBuffer','on');

limiar=0.45; % Valor usado para classificar e validar a saida da rede
             % Limiar usado para declarar MISCLASSIFICATION
             % considerando saidas no intervalo [0,1]
             
hexastr2=['0123456789ABCDEFX'];
Str2=[' 0123456789ABCDEF ']';
nsizegrid=7;
npat=16;
colormap(gray(2)); map=colormap;
% map(1,:) = black, map(2,:) = branco
% black = [0 0 0]; branco=[1 1 1];
map(1,:)=0.5*[1 1 1];
map(1,:)=[0 1 0]; % green
map(1,:)=[1 1 1]; % branco
map(2,:)=[1 0.675 0]; % amarelo
figure(Nf);
colormap(map); axis('square');
imgtest=2*ones(nsizegrid,nsizegrid);
imgtest=ones(nsizegrid,nsizegrid);
image(imgtest); drawnow; pause(1);

disp('Teste de uma rede neural com padro definido com mouse.');
disp('Modifique o padro de entrada selecionado usando o mouse.');
disp('Para trocar de padro de entrada, clicle fora da figura.');
disp('Para interromper o teste, digite X ...');
disp(' ');
hexapat;

padraoref=' ';
while strcmp(padraoref,'X') == 0;
      while length(findstr(padraoref,hexastr2)) == 0;
            padraoref=input('Selecione padro de referncia (X p/ fim) = ','s');
            padraoref=upper(padraoref);
            if length(findstr(padraoref,hexastr2)) == 0 | length(padraoref) > 1;
               fprintf('  Padro selecionado invlido ...\n');
            end;
      end;
      % Neste ponto o padrao de referencia selecionado foi julgado valido (0-9,A-F,X)
      if strcmp(padraoref,'X') == 0; % se nao foi selecionado FIM continue
         eval(['imgtest=INPD' padraoref ';']); % carregue o padrao selecionado
         nTestPat=0; % indica numero de testes neste carregamento deste padrao
         % pixel value 0 --> cor(2), pixel value 1 --> cor(1)
         Vcontinua='S';
         while strcmp(Vcontinua,'S')==1;
               figure(Nf);
               image(2-imgtest);
               title(['Padro de referncia = ' padraoref]);
               drawnow;
               % Agora pegue a entrada do mouse
               indmouse=0; % indmouse = 1 se mouse clicado fora da figura
               while indmouse==0;
                     % pegar o ponto onde o mouse foi clicado
                     % exceto na vez que este padrao foi carregado
                     figure(Nf);
                     Xmouse=1; Ymouse=1;
                     if nTestPat > 0; [Xmouse,Ymouse]=ginput(1); end;
                     % Verifica se mouse foi clicado fora da figura, no more points
                     if Xmouse < 0.5 | Xmouse > 7.5 | Ymouse < 0.5 | Ymouse > 7.5;
                        indmouse=1; % indicador de mouse clicado fora da figura
                     else;
                        % Mouse clicado dentro da figura.
                        % Entao calcular a saida da rede neural
                        Xmouse=round(Xmouse); Ymouse=round(Ymouse);
                        % Xmouse = numero da coluna; Ymouse = numero da linha
                        % inverter a cor deste ponto
                        % Pixel value: 0 --> 1, 1 --> 0
                        if nTestPat > 0;
                          imgtest(Ymouse,Xmouse)=1-imgtest(Ymouse,Xmouse);
                        end;
                        % pixel value 0 --> cor(2), pixel value 1 --> cor(1)
                        % Mostre o padrao alterado pelo mouse
                        figure(Nf);
                        image(2-imgtest);
                        title(['Padro de referncia = ' padraoref]);
                        drawnow; refresh;
                        % Agora finalmente teste este padrao
                        % a) Empilhe as linhas e forme uma coluna de (nsizegrid)^2 elementos
                        for iemp=1:nsizegrid;
                            Inptest((iemp-1)*nsizegrid+1:nsizegrid*iemp,1)=imgtest(iemp,:)';
                        end;
                        % Escalona o vetor de entrada da rede
                        Inptest=2*Inptest-1;  
                        % Calcula a saida da rede neural de KOHONEN (Outtest)
                        % 1. Calcula as coordenadas da unidade vencedora para este padrao de teste
                        [iv,jv]=khout(W,Inptest,N1,N2);
                        if exist('htk'); delete(htk); end;
                        figure(Nf-2); htk=text(jv+0.1,iv,'@');
                        set(htk,'Color',[0 0 0]);
                        % 2. Calcula a distancia entre unidade vencedora e as unidades
                        % vencedoras para cada padrao sem ruido ao final do treinamento.
                        % As distancias sao entao normalizadas pela maior distancia.
                        for ipat=1:Npat;
                            d(ipat)=sum(abs(UV(ipat,:)-[iv,jv]));
                        end;
                        d=d/max(d); 
                        % 3. Assumindo que o grau de pertinencia ao cluster varia exponencialmente
                        % com a distancia normalizada para as unidades vencedoras
                        % para cada padrao sem ruido ao final do treinamento.
                        % Assim, se dist = 0, grau de pertinencia = 1
                        %        se dist = 1, grau de pertinencia bem baixo
                        % No pior caso, d=dmin=1/(N1+N2).
                        % Assim Outtest para dmin deve ser menor que limiar.
                        % P. ex., exp(-1)=0,368. Assim podemos escolher
                        % a funcao de saida como: exp(-(N1+N2)*dmin) = exp(-1) < limiar
                        Outtest=exp(-(N1+N2)*d);
                        % Neste ponto saida da rede neural entre 0 e 1
                        nout=length(Outtest);
                        % Mostra saidas da rede neural
                        figure(Nf+1);
                        for iout=1:nout;
                            plot([iout iout],[0 Outtest(iout)]); hold on;
                        end;
                        plot([0 nout+1],[limiar limiar],'--r');
                        plot([0 nout+1],[1-limiar 1-limiar],'--r');
                        hold off; axis([0 nout+1 0 1]);
                        xlabel('Classes na camada de sada');
                        ylabel('Valor da sada');
                        set(gca,'Xtick',[0:nout]);
                        set(gca,'XTickLabel',Str2);
                        title('Sada da Rede Neural');
                        drawnow; refresh;
                        % Verifica validade da resposta da rede
                        % Aplica limiar na vetor resposta da rede
                        for iout=1:nout;
                            if (1-limiar) < Outtest(iout); Outtest(iout)=1; end;
                            if Outtest(iout) < limiar;     Outtest(iout)=0; end;
                            if limiar <= Outtest(iout) & Outtest(iout) <= (1-limiar);
                               Outtest(iout)=0.5;
                            end;
                        end;
                        % Verifica validade da resposta da rede neural
                        if sum(Outtest>(1-limiar))~=1 | sum(Outtest<limiar)~=(nout-1);
                           % Resposta invalida
                           fprintf('==> Resposta da rede neural = INVALIDA !!!\n    ');
                           figure(Nf+1); text(13,0.8,['==> ' hexastr2(17)]);
                           % Assumindo que neste ponto a resposta da rede e': 0, 0.5 ou 1
                           for ipat=1:npat;
                               if Outtest(ipat)~=0.5;
                                  fprintf('%s ',hexastr2(ipat));
                               else;
                                  fprintf(' %s  ',hexastr2(ipat));
                               end;
                           end;
                           fprintf('\n    ');
                           fprintf('%g ',Outtest);
                           fprintf('\n');
                        else;
                           % Resposta valida
                           [Outnnmax,Ioutnnmax]=max(Outtest);
                           disp(['==> Resposta da rede neural = ', hexastr2(Ioutnnmax)]);
                           figure(Nf+1); text(13,0.8,['==> ' hexastr2(Ioutnnmax)]);
                        end;
                        nTestPat=nTestPat+1; % incrementa o no. de testes
                                             %neste carregamento deste padrao
                     end; % end for "if Xmouse < 0.5 | Ymouse ..."
               end;  % end for "while indmouse==0"
               % Decidir acao caso mouse tenha sido clicado fora da figura   
               % Testar se deseja-se continuar com mesmo padrao definido ate' aqui,
               % o qual possivelmente ja foi alterado com o mouse
               disp(' ');
               Vcontinua=input('Continuar com esta imagem (S/N) = ','s');
               Vcontinua=upper(Vcontinua);
         end; % end for "while strcmp(Vcontinua,'S')==1;" 
         % reset padraoref para mostrar outro padrao
         padraoref=' ';
      else;
         fprintf('Bye bye ... \n\n');
      end; % end for "if strcmp(padraoref,'X') == 0;"
end; % end for "while strcmp(padraoref,'X') == 0"

% EOF TPNN2.M
