% MATLAB script 5.3 - file: UNSUP1.M
% primeira demonstracao do principio do aprendizado nao supervisionado
% usando competicao. Rede neural com 2 entradas e 4 unidades.
% Durante o treinamento, os pesos das unidades sao normalizados.
% Exemplo da pag. 157 (fig. 12.1) do livro IA_CONTAUT, Cairo e Takashi, 2000.
% Autor: Cairo L. Nascimento Jr.(ITA), cairo@ita.br
% Data: 14/Mar/1998, Revisao em 7/Fev/2002

clear ll; close all; Nf=1; figure(Nf); set(gcf,'DoubleBuffer','on');

% 1 - Definicao dos parametros de simulacao
Wang(1,:)=[0 90 180 -90];
Xang(1,:)=[50 150 -80 30 145 40 -60 140];
neta0=0.8;
netdec=0.9;  % < 1 mas proximo de 1, p. ex., 0.95
Niter=20;

% 2 - Inicializacao dos pesos (W) e dos padroes de entrada (X)
% 2.1 - cada linha de W representa os pesos recebidos por cada unidade
Wl=length(Wang);
[W1,W2]=pol2cart(Wang*pi/180,ones(1,Wl));
W=[W1' W2'];

% 2.2 - cada coluna de X representa um vetor de treinamento
% Xang = angulos de X em graus
Xl=length(Xang);
[X1,X2]=pol2cart(Xang*pi/180,ones(1,Xl));
X=[X1; X2];

% 2.3 - plote W inicial e X
for i=1:Wl;
    hw(i)=plot(0.85*W(i,1),0.85*W(i,2),'o'); hold on;
    plot([0 0.85*W(i,1)],[0 0.85*W(i,2)],'--');
    ht(i)=text(1.1*W(i,1),1.1*W(i,2),['W' num2str(i)]);
    set(ht(i),'HorizontalAlignment','center');
end;
plot(X(1,:),X(2,:),'+');
xlabel('x1'); ylabel('x2');
title('Iterao 0');
axis([-1.2 1.2 -1.2 1.2]); axis('square');
disp('Pause on ...'); pause; disp('Pause off ...'); 

% 3 - Loop de treinamento
neta=neta0/netdec;
for iter=1:Niter;
    neta=neta*netdec;
    if iter > 5; pause off; end;
    for ipat=1:Xl;
        % indicacao da entrada
        title(['Iterao No ' num2str(iter) ', Entrada ' num2str(ipat)]);
        hx=plot(X(1,ipat),X(2,ipat),'+r');
        % 3.1 - calculo das saidas das unidades
        Y=W*X(:,ipat);
        % 3.2 - identifique e marque a unidade vencedora
        [Ymax,Imax]=max(Y);
        WImax_old=W(Imax,:);
        hp(1)=plot(0.85*W(Imax,1),0.85*W(Imax,2),'o');
        % 3.3 - atualize a unidade vencedora
        W(Imax,:)=W(Imax,:)+neta*(X(:,ipat)'-W(Imax,:));
        W(Imax,:)=W(Imax,:)/sqrt(W(Imax,:)*W(Imax,:)');
        % 3.4 - plote o novo vetor de pesos da unidade vencedora
        hp(2)=plot(0.85*W(Imax,1),0.85*W(Imax,2),'or');
        delete(ht(Imax)); % reposicione o label da unidade vencedora
        ht(Imax)=text(1.1*W(Imax,1),1.1*W(Imax,2),['W' num2str(Imax)]);
        set(ht(Imax),'HorizontalAlignment','center');
        pause;
        delete(hx); delete(hp);
        plot(0.85*WImax_old(1),0.85*WImax_old(2),'oy');
        plot(0.85*W(Imax,1),0.85*W(Imax,2),'o');
        drawnow;
    end;
    figure(Nf); drawnow; 
end;

% 4 - Plote com retas os pesos finais
for i=1:Wl;
    plot([0 0.85*W(i,1)],[0 0.85*W(i,2)],'-');
end;
figure(Nf); drawnow; 
[Wangf,Wmagf]=cart2pol(W(:,1),W(:,2));
Wangf=Wangf*180/pi;
Wangh(:,1)=[1:Wl]';
Wangh(:,2)=Wang';
Wangh(:,3)=Wangf,

% EOF unsup1.m
