% MATLAB script 5.3 - file: UNSUP2.M
% Segunda demonstracao do principio do aprendizado nao supervisionado
% usando competicao. Neste exemplo, os pesos das unidades nao sao normalizados
% durante o treinamento. Rede neural com 2 entradas e 6 unidades
% Exemplos das pags. 158 e 159 (figs. 12.2 a 12.4) do livro IA_CONTAUT,
% Cairo e Takashi, 2000.
% Autor: Cairo L. Nascimento Jr.(ITA), cairo@ita.br
% Data: 14/Mar/1998, Revisao em 7/Fev/2002

clear all; close all; Nf=1; figure(Nf);
set(gcf,'DoubleBuffer','on'); clf;
pause on;

% 1 - Definicao dos parametros de simulacao
% Os pontos de entrada sao espalhados em clusters
% que tem uma forma circular.
% Para cada cluster de pontos de entrada define-se:
% Centro (C), raio (R) e numero de pontos (Np)
% Cada cluster apresenta Np pontos de entrada uniformemente
% distribuidos dentro da regiao do cluster.
C=[ 6 -4 -4  7;
    4  4 -6 -2]; % Coluna de C = Centro de cada regiao
R=[ 2  4  4  2];
Np=[80 80 80 80];
% os pesos iniciais tem modulo 1
Wang(1,:)=[0 45 90 180 -90 -45]; % angulos dos pesos iniciais
neta0=0.01;  % Use 0.01 para fig. 12.2 e 12.3 e 0.02 para fig. 12.4
netdec=0.95; % < 1 mas proximo de 1, p. ex., 0.95
Niter=100;
Ptr=0.30;  % porcentagem do numero total de dados de entrada
           % escolhidos aleatoriamente para o treinamento
           % em cada iteracao; 0 < Ptr <= 1
rand('seed',1); % semente incial para gerar os pontos de entrada

% 2 - Inicializacao dos padroes de entrada (X) e dos pesos (W)
% 2.1 - geracao dos dados de entrada X para treinamento
Nc=length(Np); % Numero de clusters
X=[];
for ic=1:Nc;
    Ci=C(:,ic); Ri=R(ic); Npi=Np(ic);
    for ip=1:Npi;
        Xnew=2*Ri*[rand-0.5; rand-0.5];
        while Xnew'*Xnew>Ri*Ri;
              Xnew=2*Ri*[rand-0.5; rand-0.5];
        end;
        X=[X Ci+Xnew];
    end;
end;
hplot=plot(X(1,:),X(2,:),'+');
grid on; hold on;

% 2.2 - inicializacao dos pesos W
% cada linha de W representa os pesos recebidos por cada unidade
Wl=length(Wang);
[W1,W2]=pol2cart(Wang*pi/180,ones(1,Wl));
W=[W1' W2'];
plot(W(:,1),W(:,2),'.');
xlabel('Entrada x1'); ylabel('Entrada x2');
title('Iterao 0');
axis([-10 10 -10 10]); axis('square'); hold on;
disp('Pause on ...'); pause;
disp('Pause off ...');

% 3 - Loop de treinamento
neta=neta0/netdec;
Xl=size(X,2); % numero de pontos de treinamento
Nptr=ceil(Ptr*Xl);
W1=W(:,1);
W2=W(:,2);
Iw=zeros(Wl,1);  % Iw indica o numero de vezes que um peso foi atualizado
for iter=1:Niter;
    neta=neta*netdec;
    % 3.1 - selecao dos pontos de treinamento nesta iteracao
    Xt=[];
    for ix=1:Nptr;
        Xtnew=X(:,ceil(rand*Xl));
        Xt=[Xt Xtnew];
    end;
    %figure(Nf+1); clf;
    title(['Iterao No ' num2str(iter)]);
    %plot(Xt(1,:),Xt(2,:),'+'); grid;
    %axis([-10 10 -10 10]); hold on; drawnow;
    for ipat=1:Nptr;
        % 3.2 - calculo das saidas das unidades
        Y=W*Xt(:,ipat);
        % 3.3 - identifique a unidade vencedora
        [Ymax,Imax]=max(Y);
        % 3.4 - atualize a unidade vencedora
        Wold=W(Imax,:);
        W(Imax,:)=W(Imax,:)+neta*(Xt(:,ipat)'-W(Imax,:));
        % 3.5 - armazene a historia de W
        Iw(Imax)=Iw(Imax)+1;
        %W(Imax,:)=W(Imax,:)/sqrt(W(Imax,:)*W(Imax,:)');
        plot([Wold(1) W(Imax,1)],[Wold(2) W(Imax,2)],'-r');
    end;
    if 10*fix(iter/10)==iter | iter <= 10; drawnow; end;
end;

% 4 - Rotulacao - Labelling
delete(hplot); % delete and redraw the training points using the correct mark
Npc=zeros(Wl,1);
Str1=['+x.*o-'];
if neta0>=0.02; Str1=['+x*o.-']; end;
for ipat=1:Xl;
    Y=W*X(:,ipat);
    [Ymax,Imax]=max(Y);
    Npc(Imax,1)=Npc(Imax,1)+1;
    s=num2str(Imax);
    %text(X(1,ipat),X(2,ipat),s); hold on;
    plot(X(1,ipat),X(2,ipat),[Str1(Imax) 'b']); hold on;
end;
Npc, W,
grid on;
xlabel('Entrada x1'); ylabel('Entrada x2');
title('Rotulao');
axis([-10 10 -10 10]); axis('square'); drawnow;

% EOF unsup2.m
